import os
import subprocess
import logging
from ga2.common.cmdExecuter import CmdExecuter

logger = logging.getLogger(__name__)


class IProxyManager(object):

    def __init__(self, udid):
        self.udid = udid
        self.forward_list = {}

    def forward(self, localport, remoteport):
        localport = str(localport)
        remoteport = str(remoteport)
        cmd = "iproxy " + localport + " " + remoteport
        if self.udid:
            cmd += " " + self.udid
        CmdExecuter.execute_no_wait(cmd)
        self.forward_list[localport] = remoteport

    def remove_forward(self, localport):
        localport = str(localport)
        cmd = "iproxy " + localport + " "
        CmdExecuter.execute_and_wait("pkill -f \"" + cmd + "\"")
        if localport in self.forward_list.keys():
            self.forward_list.pop(localport)

    def remove_all_forwards(self):
        for key, value in self.forward_list.items():
            localport = str(key)
            cmd = "iproxy " + localport + " "
            CmdExecuter.execute_and_wait("pkill -f \"" + cmd + "\"")
        self.forward_list.clear()

    # @staticmethod
    # def _init_wdaServer():
    #     udid = os.environ.get("IOS_SERIAL")
    #     if udid is None:
    #         process=subprocess.Popen("idevice"
    #     mXcodeCmd =
    #         "xcodebuild -project %s/WebDriverAgent.xcodeproj test -scheme WebDriverAgentRunner -destination 'platform=iOS,id=%s'",
    #         wdaDriverPath, udid);



# -*- coding: UTF-8 -*-


class UIElement():
    """
        in unity game Element is just GameObject
    Attributes:
        object_name:GameObject的全路径
        instance:GameObject Instance,the instance id of gameobject is always guarnnteed to be unique
    """

    def __init__(self, element,bound):
        self.__element = element
        self.__bound = bound

    @property
    def element(self):
        return self.__element

    @property
    def bound(self):
        return self.__bound

    def __str__(self):
        return "element {0} bound = {1}".format(self.element, self.bound)

    def __eq__(self, uielement):
        return hasattr(uielement, 'element') and self.element.instance == uielement.element.instance

    def __ne__(self, uielement):
        return not self.__eq__(uielement)

    def __repr__(self):
        return '<{0.__module__}.{0.__name__} (object_name="{1}", instance="{2}", bound="{3}")>'.format(type(self), self.element.object_name, self.element.instance,self.element.bound)




from ga2.common.loggerConfig import logger
from ga2.common.utils import *
from ga2.cloud.reporter import Reporter
from ga2.common.WebDriverWait import WebDriverWait


class AutomationHelperEngine:

    @staticmethod
    def wait_element_by_name(device, name, timeout):
        if device and device.engine_connector():
            element = None
            try:
                element = WebDriverWait(timeout, 2).until(device.engine_connector().find_element, name)
                print(element)
            except Exception as e:
                logger.warn("element wait timeout:" + name)
            return element
        return None

    @staticmethod
    def wait_element_by_path(device, path, timeout):
        if device and device.engine_connector():
            elements = None
            try:
                elements = WebDriverWait(timeout, 2).until(device.engine_connector().find_element_path, path)
                if len(elements) > 0:
                    return elements[0]
            except Exception as e:
                logger.warn("element wait timeout:" + path)
        return None

    @staticmethod
    def wait_elements_by_path(device, path, timeout):
        if device and device.engine_connector():
            elements = None
            try:
                elements = WebDriverWait(timeout, 2).until(device.engine_connector().find_elements_path, path)
                print(elements)
            except Exception as e:
                logger.warn("element wait timeout:" + path)
            return elements
        return None

    @staticmethod
    def handle_element_by_name(device, name, operator, **kwargs):
        if not device or not device.engine_connector():
            return None
        engine = device.engine_connector()
        element = engine.find_element(name)
        print(element)
        if element is None:
            logger.error("engine element is not found : " + name)
            return None
        bound = engine.get_element_bound(element)
        if bound is None:
            logger.error("engine element is not found : " + name)
            return None
        target_pos = (bound.x + bound.width / 2, bound.y + bound.height / 2)
        if is_cloud_mode():
            (width, height) = device.display_size()
            Reporter().screenshot_with_mark(width, height, target_pos[0], target_pos[1])
        operator(target_pos[0], target_pos[1], **kwargs)
        return target_pos

    @staticmethod
    def handle_element_by_self(device, element, operator, **kwargs):
        if not device or not device.engine_connector():
            return None
        engine = device.engine_connector()
        if element is None:
            logger.error("engine element is none")
            return None
        bound = engine.get_element_bound(element)
        if bound is None:
            logger.error("engine element is not found... ")
            return None
        target_pos = (bound.x + bound.width / 2, bound.y + bound.height / 2)
        if is_cloud_mode():
            (width, height) = device.display_size()
            Reporter().screenshot_with_mark(width, height, target_pos[0], target_pos[1])
        operator(target_pos[0], target_pos[1], **kwargs)
        return target_pos

    @staticmethod
    def handle_element_by_path(device, path, operator, **kwargs):
        if not device or not device.engine_connector():
            return None
        engine = device.engine_connector()
        elements = engine.find_elements_path(path)
        if elements is None or len(elements) == 0:
            logger.error("engine element is not found : " + path)
            return None
        if len(elements) > 1:
            logger.warn("multiple engine element is mathed by path : " + path)
        element = elements[0]
        bound = engine.get_element_bound(element)
        target_pos = (bound.x + bound.width / 2, bound.y + bound.height / 2)
        if is_cloud_mode():
            (width, height) = device.display_size()
            Reporter().screenshot_with_mark(width, height, target_pos[0], target_pos[1])
        operator(target_pos[0], target_pos[1], **kwargs)
        return target_pos

    @staticmethod
    def get_bound_by_name(device, name):
        if not device or not device.engine_connector():
            return None
        engine = device.engine_connector()
        element = engine.find_element(name)
        if element is None:
            logger.error("engine element is none")
            return None
        bound = engine.get_element_bound(element)
        if bound is None:
            logger.error("engine element is not found... ")
            return None
        return (bound.x, bound.y, bound.x + bound.width, bound.y + bound.height)

    @staticmethod
    def get_bound_by_path(device, path):
        if not device or not device.engine_connector():
            return None
        engine = device.engine_connector()
        elements = engine.find_elements_path(path)
        if elements is None or len(elements) == 0:
            logger.error("engine element is not found : " + path)
            return None
        if len(elements) > 1:
            logger.warn("multiple engine element is mathed by path : " + path)
        element = elements[0]
        if element is None:
            logger.error("engine element is none")
            return None
        bound = engine.get_element_bound(element)
        if bound is None:
            logger.error("engine element is not found... ")
            return None
        return (bound.x, bound.y, bound.x + bound.width, bound.y + bound.height)

    @staticmethod
    def get_bound_by_self(device, element):
        if not device or not device.engine_connector():
            return None
        engine = device.engine_connector()
        if element is None:
            logger.error("engine element is none")
            return None
        bound = engine.get_element_bound(element)
        if bound is None:
            logger.error("engine element is not found... ")
            return None

        return (bound.x, bound.y, bound.x + bound.width, bound.y + bound.height)



# -*- coding: UTF-8 -*-

import os
import logging
import subprocess
import time
import re
logger=logging.getLogger("wetest")



def adb_wait(cmd, serial=None):
    logger.debug(cmd)
    if serial:
        command = "adb -s {0} {1}".format(serial, cmd)
    else:
        command = "adb {0}".format(cmd)
    p = subprocess.Popen(command, shell=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
    out, err = p.communicate()
    if err is not None and len(err)>0:
        logger.error("[adb_wait] " + cmd + " error: " + str(err))
    return out, err

def adb_push(local_path, target_path,serial):
    cmd = "push {0} {1}".format(local_path, target_path )
    return adb_wait(cmd, serial)

def adb_pull(remote_path, target_path,serial):
    cmd = "pull {0} {1}".format(remote_path, target_path )
    return adb_wait(cmd, serial)

def adb_forward(pc_port, mobile_port,serial):
    cmd = "forward tcp:{0} tcp:{1}".format(pc_port, mobile_port)
    return adb_wait(cmd, serial)

def adb_remove_forward(pc_port,serial):
    cmd = "forward --remove tcp:{0} ".format(pc_port)
    return adb_wait(cmd, serial)

'''
if needStdout is true , the caller must read the process pipe out in case the subprocess hangs
'''
def adb_nowait(cmd, shell=False, serial=None, needStdout=True):
    if serial:
        command = "adb -s {0} {1}".format(serial, cmd)
    else:
        command = "adb {0}".format(cmd)
    #print "daemon: {0}".format(command)
    if needStdout is True:
        p = subprocess.Popen(command, shell=shell, stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
    else:
        p = subprocess.Popen(command, shell=shell)
    return p

def adb_kill_process_by_name(name, serial=None):
    pid = adb_get_pid_by_name(name,serial)
    logger.debug("[kill_process_by_name] : " + name + " pid : " + str(pid))
    if pid is not None and pid > 0:
        adb_wait("shell kill {0}".format(pid), serial)
    time.sleep(1)
    pid = adb_get_pid_by_name(name, serial)
    if pid is not None and pid > 0:
        adb_wait("shell kill -9 {0}".format(pid), serial)

def adb_get_pid_by_name(name, serial=None):
    logger.debug("[get_pid_by_name] : " + name)
    out, err = adb_wait("shell ps", serial)
    # outstr = "\n".join(out)
    pattern = "\w+\s+(\d+)\s+.+" + name
    match = re.search(pattern, out.decode("utf-8"), re.M)
    if match:
        pid = match.group(1)
        logger.debug("found process pid= {0}".format(pid))
        return int(pid)
    else:
        out, err = adb_wait("shell ps -ef", serial)
        # print(out)
        # outstr = "\n".join(out)
        match = re.search(pattern,  out.decode("utf-8"), re.M)
        if match:
            pid = match.group(1)
            logger.debug("found process pid= {0}".format(pid))
            return int(pid)
    return 0


def adb_get_top_app(serial=None):
    out, err = adb_wait("shell dumpsys window windows", serial)
    pattern = "mCurrentFocus=.*\s([^\s]*)/(.*?)}?\s"
    package = None
    match = re.search(pattern, out.decode("utf-8"), re.M)
    if match:
        package = match.group(1)
        logger.debug("found top package by adb = {0}".format(package))
    return package

def adb_get_display_size(serial=None):
    out, err = adb_wait("shell dumpsys window displays", serial)
    pattern = "cur=(\d+)x(\d+)"
    display_size = None
    match = re.search(pattern, out.decode("utf-8"), re.M)
    if match:
        display_size = (int(match.group(1)), int(match.group(2)))

    return display_size

if __name__ == "__main__":
    adb_wait("shell input tap 553 1007")
    pass



# -*- coding: UTF-8 -*-
import traceback
import time
import threading
import logging
import os

logger = logging.getLogger("gautomator")


def is_cloud_mode():
    return "PLATFORM_PORT" in os.environ and os.environ.get("PLATFORM_PORT") is not None and os.environ.get("PLATFORM_PORT") != ""


isInCloudMode = is_cloud_mode


def callLog(func):
    def wrapper(*args, **kw):
        local_time = time.time()
        ret=func(*args, **kw)
        logger.debug('Call [%s] cost: %.2fs' % (func.__name__ ,time.time() - local_time))
        return ret
    return wrapper


class ErrType:
    ERR_SUCCEED = 0
    ERR_TIMEOUT = -101
    ERR_CONNECT_TO_SDK_FAILED = -102
    ERR_CONNECT_TO_UITUAOMTTOR_FAILED = -103
    ERR_LOGIN_FAILED = -104
    ERR_WDA_NOT_RUNNING = -105
    ERR_DEVICE_NOT_INITED = -106
    ERR_CLOUD_PLATFORM = -107
    ERR_INVALID_ENGINETYPE = -108
    ERR_PROCESS_NOT_FOUND = -109
    ERR_UNIMPLEMENTED = -110
    ERR_ADB = -111





# The contents of this file has been derived code from the Twisted project
# (http://twistedmatrix.com/). The original author is Jp Calderone.

# Twisted project license follows:

# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import unicode_literals

from six import binary_type, text_type, byte2int, iterbytes, unichr


PRINTABLE = set(range(0x20, 0x26)) | set(range(0x27, 0x7f))

# TODO: module needs refactoring (e.g. variable names suck)


def encode(s):
    """Encode a folder name using IMAP modified UTF-7 encoding.

    Input is unicode; output is bytes (Python 3) or str (Python 2). If
    non-unicode input is provided, the input is returned unchanged.
    """
    if not isinstance(s, text_type):
        return s

    r = []
    _in = []

    def extend_result_if_chars_buffered():
        if _in:
            r.extend([b'&', modified_utf7(''.join(_in)), b'-'])
            del _in[:]

    for c in s:
        if ord(c) in PRINTABLE:
            extend_result_if_chars_buffered()
            r.append(c.encode('latin-1'))
        elif c == '&':
            extend_result_if_chars_buffered()
            r.append(b'&-')
        else:
            _in.append(c)

    extend_result_if_chars_buffered()

    return b''.join(r)


AMPERSAND_ORD = byte2int(b'&')
DASH_ORD = byte2int(b'-')


def decode(s):
    """Decode a folder name from IMAP modified UTF-7 encoding to unicode.

    Input is bytes (Python 3) or str (Python 2); output is always
    unicode. If non-bytes/str input is provided, the input is returned
    unchanged.
    """

    if not isinstance(s, binary_type):
        return s

    r = []
    _in = bytearray()
    for c in iterbytes(s):
        if c == AMPERSAND_ORD and not _in:
            _in.append(c)
        elif c == DASH_ORD and _in:
            if len(_in) == 1:
                r.append('&')
            else:
                r.append(modified_deutf7(_in[1:]))
            _in = bytearray()
        elif _in:
            _in.append(c)
        else:
            r.append(unichr(c))
    if _in:
        r.append(modified_deutf7(_in[1:]))
    return ''.join(r)


def modified_utf7(s):
    s_utf7 = s.encode('utf-7')
    return s_utf7[1:-1].replace(b'/', b',')


def modified_deutf7(s):
    s_utf7 = b'+' + s.replace(b',', b'/') + b'-'
    return s_utf7.decode('utf-7')



# -*- coding: UTF-8 -*-
"""
Tencent is pleased to support the open source community by making GAutomator available.
Copyright (C) 2016 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.

"""
__author__ = 'minhuaxu wukenaihesos@gmail.com,alexkan kanchuanqi@gmail.com'


class Commands(object):
    GET_VERSION = 100  # 获取版本号
    FIND_ELEMENTS = 101  # 查找节点
    FIND_ELEMENT_PATH = 102  # 模糊查找
    GET_ELEMENTS_BOUND = 103  # 获取节点的位置信息
    GET_ELEMENT_WORLD_BOUND = 104  # 获取节点的世界坐标
    GET_UI_INTERACT_STATUS = 105  # 获取游戏的可点击信息，包括scene、可点击节点，及位置信息
    GET_CURRENT_SCENE = 106  # 获取Unity的Scene名称
    GET_ELEMENT_TEXT = 107  # 获取节点的文字内容
    GET_ELEMENT_IMAGE = 108  # 获取节点的图片名称
    GET_REGISTERED_HANDLERS = 109  # 获取注册的函数的名称
    CALL_REGISTER_HANDLER = 110  # 调用注册的函数
    SET_INPUT_TEXT = 111  # input控件更换文字信息
    GET_OBJECT_FIELD=112 # 通过反射获取gameobject中component的属性值
    FIND_ELEMENTS_COMPONENT=113 #获取所有包含改组件的gameobject
    SET_CAMERA_NAME=114 #设置渲染的最佳的Camera
    GET_COMPONENT_METHODS = 115  # 反射获取组件上的方法
    CALL_COMPONENT_MOTHOD = 116  # 通过反射调用组件的函数
    LOAD_TEST_LIB=117 #初始化testlib服务

    PRC_SET_METHOD=118#注册python端的方法
    RPC_METHOD = 119#游戏内的接口可调用，python端的方法
    SET_SERVER_TIME = 120  # 设置服务器时间

    #######################/
    HANDLE_TOUCH_EVENTS = 200  # 发送down move up

    DUMP_TREE = 300

    #######################/
    JUMP_BUILDING = 500       # 跳转到某些建筑上
    GET_POS_TEXT = 501        # 获取坐标点的文字
    SET_INPUT_TEXT_VC = 502   # 设置input空间文字

class TouchEvent(object):
    ACTION_DOWN = 0
    ACTION_UP = 1
    ACTION_MOVE = 2

    def __init__(self, x, y, sleeptime, type):
        self.x = x
        self.y = y
        self.sleeptime = sleeptime
        self.type = type



from ga2.automation.by import By
from ga2.common.loggerConfig import logger
from ga2.common.utils import callLog
from ga2.device.device import Device
from ga2.common.utils import *
from ga2.cloud.reporter import Reporter
from ga2.automation.uiconfig import UIConfig
from ga2.common.WebDriverWait import WebDriverWait
from ga2.automation.automationHelperEngine import AutomationHelperEngine
from ga2.automation.automationHelperNative import AutomationHelperNative
from ga2.automation.automationHelperImage import AutomationHelperImage


class AutomationHelper:

    touch_method_map = {By.NAME_IN_ENGINE: AutomationHelperEngine.handle_element_by_name,
                        By.PATH_IN_ENGINE: AutomationHelperEngine.handle_element_by_path,
                        By.TEXT_NATIVE: AutomationHelperNative.handle_element_by_text,
                        By.CONDITIONS_NATIVE: AutomationHelperNative.handle_element_by_conditions,
                        By.IMAGE_TEMPLATE: AutomationHelperImage.handle_element_by_image_template,
                        By.ELEMENT_IN_ENGINE: AutomationHelperEngine.handle_element_by_self}
    # double_touch_method_map = {By.NAME_IN_ENGINE: AutomationHelperEngine.handle_element_by_name,
    #                            By.PATH_IN_ENGINE: AutomationHelperEngine.handle_element_by_path,
    #                            By.TEXT_NATIVE: AutomationHelperNative.handle_element_by_text,
    #                            By.CONDITIONS_NATIVE: AutomationHelperNative.handle_element_by_conditions,
    #                            By.IMAGE_TEMPLATE: AutomationHelperImage.handle_element_by_image_template,
    #                            By.ELEMENT_IN_ENGINE: AutomationHelperEngine.handle_element_by_self}
    long_press_method_map = {By.NAME_IN_ENGINE: AutomationHelperEngine.handle_element_by_name,
                             By.PATH_IN_ENGINE: AutomationHelperEngine.handle_element_by_path,
                             By.TEXT_NATIVE: AutomationHelperNative.handle_element_by_text,
                             By.CONDITIONS_NATIVE: AutomationHelperNative.handle_element_by_conditions,
                             By.IMAGE_TEMPLATE: AutomationHelperImage.handle_element_by_image_template,
                             By.ELEMENT_IN_ENGINE: AutomationHelperEngine.handle_element_by_self}
    wait_method_map = {By.NAME_IN_ENGINE: AutomationHelperEngine.wait_element_by_name,
                       By.PATH_IN_ENGINE: AutomationHelperEngine.wait_element_by_path,
                       By.TEXT_NATIVE: AutomationHelperNative.wait_element_by_text,
                       By.CONDITIONS_NATIVE: AutomationHelperNative.wait_element_by_conditions,
                       By.IMAGE_TEMPLATE: AutomationHelperImage.wait_element_by_image_template
                       }

    get_bound_map = {By.NAME_IN_ENGINE: AutomationHelperEngine.get_bound_by_name,
                       By.PATH_IN_ENGINE: AutomationHelperEngine.get_bound_by_path,
                       By.ELEMENT_IN_ENGINE: AutomationHelperEngine.get_bound_by_self,
                       By.TEXT_NATIVE: AutomationHelperNative.get_bound_by_text,
                       By.CONDITIONS_NATIVE: AutomationHelperNative.get_bound_by_conditions,
                       By.ELEMENT_NATIVE: AutomationHelperNative.get_bound_by_self
                       }
    wait_list_method_map = {By.PATH_IN_ENGINE: AutomationHelperEngine.wait_elements_by_path}


    # By.ID_IN_UICONFIG: AutomationHelper.wait_element_by_uiconfig}

    @staticmethod
    @callLog
    def wait_element(method, param, timeout, device=None):
        device = device or Device.get_default()
        if not device:
            logger.error("device is not set in automationAPI")
            return None
        if method == By.ID_IN_UICONFIG:
            uiconfig = UIConfig.get_default()
            if uiconfig == None:
                return None
            (method, param) = uiconfig.get_locator(param, device.device_type())
        if method not in AutomationHelper.wait_method_map:
            logger.error("invalid find method :" + str(method))
            return None
        return AutomationHelper.wait_method_map.get(method)(device, param, timeout)

    @staticmethod
    @callLog
    def wait_elements(method, param, timeout, device=None):
        device = device or Device.get_default()
        if not device:
            logger.error("device is not set in automationAPI")
            return None
        if method == By.ID_IN_UICONFIG:
            uiconfig = UIConfig.get_default()
            if uiconfig == None:
                return None
            (method, param) = uiconfig.get_locator(param, device.device_type())
        if method not in AutomationHelper.wait_method_map:
            logger.error("invalid find method :" + str(method))
            return None
        return AutomationHelper.wait_list_method_map.get(method)(device, param, timeout)

    @staticmethod
    @callLog
    def touch_element(method, param, device=None, sleep_after=0.5):
        device = device or Device.get_default()
        if not device:
            logger.error("device is not set in automationAPI")
            return None
        if method == By.ID_IN_UICONFIG:
            uiconfig = UIConfig.get_default()
            if uiconfig == None:
                return None
            (method, param) = uiconfig.get_locator(param, device.device_type())

        if method not in AutomationHelper.touch_method_map:
            logger.error("invalid touch method :" + str(method))
            return None
        ret = AutomationHelper.touch_method_map.get(method)(device, param, device.touch)
        if ret and sleep_after and sleep_after>0:
            time.sleep(sleep_after)
        return ret

    @staticmethod
    @callLog
    def long_press_element(method, param, duration=2, device=None, sleep_after=0.5):
        device = device or Device.get_default()
        if not device:
            logger.error("device is not set in automationAPI")
            return None
        if method == By.ID_IN_UICONFIG:
            uiconfig = UIConfig.get_default()
            if uiconfig == None:
                return None
            (method, param) = uiconfig.get_locator(param, device.device_type())
        if method not in AutomationHelper.long_press_method_map:
            logger.error("invalid long_press method :" + method)
            return None
        ret = AutomationHelper.long_press_method_map.get(method)(device, param, device.long_press, duration=duration)
        if ret and sleep_after and sleep_after>0:
            time.sleep(sleep_after)
        return ret

    @staticmethod
    @callLog
    def get_element_bound(method, param,device=None):
        device = device or Device.get_default()
        if not device:
            logger.error("device is not set in automationAPI")
            return None
        if method == By.ID_IN_UICONFIG:
            uiconfig = UIConfig.get_default()
            if uiconfig == None:
                return None
            (method, param) = uiconfig.get_locator(param, device.device_type())
        if method not in AutomationHelper.get_bound_map:
            logger.error("invalid get_bound_method :" + method)
            return None
        return  AutomationHelper.get_bound_map.get(method)(device, param)


    # @staticmethod
    # @callLog
    # def double_touch_element(method, param, device=None, sleep_after=0.5):
    #     device = device or Device.get_default()
    #     if not device:
    #         logger.error("device is not set in automationAPI")
    #         return None
    #     if method == By.ID_IN_UICONFIG:
    #         uiconfig = UIConfig.get_default()
    #         if uiconfig == None:
    #             return None
    #         (method, param) = uiconfig.get_locator(param, device.device_type())
    #     if method not in AutomationHelper.double_touch_method_map:
    #         logger.error("invalid double touch touch method :" + method)
    #         return None
    #     ret = AutomationHelper.long_press_method_map.get(method)(device, param, device.double_touch)
    #     if ret and sleep_after and sleep_after > 0:
    #         time.sleep(sleep_after)
    #     return ret


# -*- coding: UTF-8 -*-
"""
Tencent is pleased to support the open source community by making GAutomator available.
Copyright (C) 2016 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.

"""

import time
import os
import logging
from ga2.cloud.httptools.remoteConnection import RemoteConnection, Method
from ga2.common.wetestExceptions import WeTestPlatormError, WeTestInvaildArg

logger = logging.getLogger(__name__)


class Command(object):
    TOUCH_CAPTURE = (Method.POST, "touchcapture")
    CAPTURE = (Method.POST, "snapshot")
    # GET_SCREEN_IMAGE = (Method.GET, "screenimage")
    FORWARD = (Method.POST, "forward")
    RESTORE_WDA = (Method.POST, "restorewda")
    LAUNCH_APP = (Method.POST, "launchapp")
    GET_ROTATION = (Method.GET, "rotation")
    CLEAR_APP_DATA = (Method.POST, "clearappdata")
    CURRENT_PACKAGE_NAME = (Method.GET, "currentpackagename")
    RESOLUTION = (Method.GET, "resolution")
    STOP_UIAUTOMATOR = (Method.GET, "pauseuiautomator")
    START_UIAUTOMATOR = (Method.GET, "resumeuiautomator")
    TOUCH = (Method.POST, "touch")
    GET_TEST_TIME = (Method.GET, "runtime")
    # SCENE_TAG = (Method.POST, "scenetag")
    REPORT_SCENE = (Method.POST, "scenereport")
    ANDROIDVERSION = (Method.GET, "androidversion")
    MODEL = (Method.GET, "model")
    REPORT_ERROR = (Method.POST, "reporterror")
    PROCDIED_REPORT = (Method.POST, "procdiedreport")
    SCENE_CAPTURE_TAG = (Method.POST, "scene_capture_tag")


class Executor(object):
    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.url = "http://{0}:{1}".format(host, port)
        # self.platform_client = RemoteConnection(url, keep_alive=True)

    def _check_response(self, response):
        """
            errorcode==0 means success
            :param response:
            :return the data in response:
        """
        if not response:
            raise WeTestPlatormError("None response")

        if isinstance(response, bool):
            return response

        if not isinstance(response, dict):
            invaild_response = "Invaild response : {0}".format(response)
            raise WeTestPlatormError(invaild_response)

        error_code = response.get("result", 0)
        if error_code == 0:
            return response.get("data", True)  # 如果没有数据，还是会返回，只是不再会有data数据
        else:
            errorMessage = response.get("errorMsg", "platorm api error")
            raise WeTestPlatormError(errorMessage)

    def excute_platform(self, command, params=None):
        """
        :param command:
        :param params:
        :return if errorcode is 0, return the data or True if data is None :
        """
        platform_client = RemoteConnection(self.url, keep_alive=False)
        start_time = time.time()
        if not isinstance(command, tuple):
            raise WeTestInvaildArg("command is invaild")

        if command[0] == Method.GET:
            response = platform_client.get(command[1], params)
        else:
            response = platform_client.post(command[1], params)
        end_time = time.time()

        if response and isinstance(response, dict) and "data" in response:
            if "imageData" in response["data"]:
                response = self._check_response(response)
                return response
        logger.debug("Command: {0} Response: {1} time: {2}s".format(command, response, (end_time - start_time)))
        response = self._check_response(response)
        return response

    # def take_screenshot(self):
    #     return self.excute_platform(Command.CAPTURE)

    def restore_wda(self, port):
        return self.excute_platform(Command.RESTORE_WDA, {"port": port})

    # def touch_capture(self, width, height, x, y, name="wetest"):
    #     return self.excute_platform(Command.TOUCH_CAPTURE,
    #                                 {"name": name, "width": width,
    #                                  "height": height, "x": x, "y": y})

    # def get_screenimage(self):
    #     return self.excute_platform(Command.GET_SCREEN_IMAGE)

    def platform_forward(self, remote_port):
        return self.excute_platform(Command.FORWARD, {"remotePort": remote_port})

    def get_top_package_activity(self):
        return self.excute_platform(Command.CURRENT_PACKAGE_NAME)

    def get_display_size(self):
        return self.excute_platform(Command.RESOLUTION)

    def get_android_version(self):
        return self.excute_platform(Command.ANDROIDVERSION)

    def get_android_model(self):
        return self.excute_platform(Command.MODEL)

    def get_rotation(self):
        return self.excute_platform(Command.GET_ROTATION)

    def clear_app_data(self, pkgname):
        return self.excute_platform(Command.CLEAR_APP_DATA, {"clearPkgName": pkgname})

    def launch_app(self, pkgname, activity="android.intent.category.LAUNCHER"):
        return self.excute_platform(Command.LAUNCH_APP, {"pkgName": pkgname, "activity": activity})

    def take_screenshot(self, sceneid=""):
        (testid, deviceid) = self.__getTestDeviceId()
        params = { "sceneid":sceneid}
        if testid > 0 and deviceid > 0 :
            params["testid"] = testid
            params["deviceid"] = deviceid
        return self.excute_platform(Command.CAPTURE,params)

    def __getTestDeviceId(self):
        testid, deviceid = 0, 0
        if ("TESTID" in os.environ):
            testid = int(os.environ.get("TESTID"))
        if ("DEVICEID" in os.environ):
            deviceid = int(os.environ.get("DEVICEID"))
        return (testid,deviceid)

    # def add_scene_tag(self, tag):
    #     return self.excute_platform(Command.SCENE_TAG, {"tagName": tag})
    def touch_capture(self, width, height, x, y, name="wetest" , sceneid=""):
        (testid, deviceid) = self.__getTestDeviceId()
        params = {"name": name, "width": width,"height": height, "x": x, "y": y, "sceneid": sceneid}
        if testid > 0 and deviceid > 0 :
            params["testid"] = testid
            params["deviceid"] = deviceid
        return self.excute_platform(Command.TOUCH_CAPTURE,params)


    def scene_capture_tag(self, scenename="", timestamp=0, sceneorder=1, scenetype=1, sceneid=""):
        (testid, deviceid) = self.__getTestDeviceId()
        params = {"scenename": scenename, "timestamp": timestamp, "sceneid": sceneid,
                  "sceneorder": sceneorder, "scenetype": scenetype}
        if testid > 0 and deviceid > 0:
            params["testid"] = testid
            params["deviceid"] = deviceid
        return self.excute_platform(Command.SCENE_CAPTURE_TAG, params)

    def touch_screen(self, width, height, x, y):
        return self.excute_platform(Command.TOUCH, {"width": width, "height": height, "x": x, "y": y})

    def pause_uiautomator(self):
        return self.excute_platform(Command.STOP_UIAUTOMATOR)

    def resume_uiautomator(self):
        return self.excute_platform(Command.START_UIAUTOMATOR)

    def get_runtime(self):
        return self.excute_platform(Command.GET_TEST_TIME)

    def report_error(self, message):
        return self.excute_platform(Command.REPORT_ERROR, {"errmsg": message})

    def procdied_report(self, is_need_report):
        return self.excute_platform(Command.PROCDIED_REPORT, {"isNeedReport": is_need_report})


def get_platform_client():
    if get_platform_client.instance:
        return get_platform_client.instance
    platform_ip = os.environ.get("PLATFORM_IP", "127.0.0.1")
    platform_port = os.environ.get("PLATFORM_PORT", "40030")
    get_platform_client.instance = Executor(platform_ip, platform_port)
    return get_platform_client.instance


get_platform_client.instance = None



# -*- coding: UTF-8 -*-

from ga2.common import imapUtf7
from ga2.device.android.adbProcess import *
import platform


class InputHeler:

    @staticmethod
    def input_char(c, serial):
        if c != " ":
            cmd = 'shell input text ' + c
            adb_wait(cmd, serial)
        else:
            cmd = 'shell input keyevent 62'
            adb_wait(cmd, serial)
        time.sleep(1)

    '''
        input string without space
    '''

    @staticmethod
    def input_word(word, serial):
        if isinstance(word, str):
            word = imapUtf7.encode(word.decode("utf-8"))
        else:
            word = imapUtf7.encode(word)

        cmd = 'shell input text ' + '\\\"' + word + '\\\"'
        if platform.system() == "Linux":
            cmd = 'shell ' + '\"' + 'input text ' + '\\\"' + word + '\\\"' + '\"'
        adb_wait(cmd, serial)

    @staticmethod
    def ime_action(ime_action_code, serial):
        cmd = "shell am broadcast -a ADB_EDITOR_CODE --ei code " + str(ime_action_code)
        adb_wait(cmd, serial)
        time.sleep(1)

    @staticmethod
    def input_text(text, need_one_by_one, serial):
        if need_one_by_one:
            for c in text:
                InputHeler.input_char(c, serial)
        else:
            cur_seg = ""
            for c in text:
                if c == ' ':  # some android device will not handle space in input
                    if len(cur_seg) > 0:
                        InputHeler.input_word(cur_seg, serial)
                        cur_seg = ""
                        time.sleep(1)
                    InputHeler.input_char(c, serial)
                else:
                    cur_seg += c
            if len(cur_seg) > 0:
                InputHeler.input_word(cur_seg, serial)
                time.sleep(1)


import sys

if __name__ == '__main__':
    if (len(sys.argv) < 2):
        print("error ... please input the chinese word you want to type in")
        exit(-1)
    InputHeler.input_text(sys.argv[1], False, "012345678F")



#-*- coding: UTF-8 -*-

import abc, six
import cv2
from ga2.device.deviceOrientation import DeviceOrientation

@six.add_metaclass(abc.ABCMeta)
class ScreenShoter():

    @abc.abstractmethod
    def screenshot(self, localpath=None, portrait=True, height=None):
        pass

    # @abc.abstractmethod
    # def screenshot_portrait(self, localpath=None, height=None):
    #     pass


    def rotate_to_portrait(self, origin_image, origin_orientation):
        if origin_orientation is DeviceOrientation.PORTRAIT:
            return origin_image
        elif origin_orientation is DeviceOrientation.LANSCAPE:
            return cv2.flip(cv2.transpose(origin_image), 1)
        elif origin_orientation is DeviceOrientation.PORTRAIT_UPSIDEDOWN:
            return cv2.flip(origin_image, -1)
        elif origin_orientation is DeviceOrientation.LANDSCAPERIGHT:
            return cv2.flip(cv2.transpose(origin_image), 0)





# -*-coding:utf-8-*-
import re
import abc, six
from ga2.common.socketClient import SocketClient
from ga2.common.wetestExceptions import *
from ga2.engine.element import Element
from ga2.engine.protocol import Commands, TouchEvent
from ga2.common.loggerConfig import logger

from enum import Enum


# class EngineType(Enum):
#     Unity=0,
#     UE4=1

class VersionInfo(object):
    """
    Attributes:
        engine_version:引擎版本信息,如5.1.0bf
        engine:Unity
        sdk_version:wetest sdk的版本信息
    """

    def __init__(self, engine_version, engine, sdk_version, ui_type):
        self.engine_version = engine_version
        self.engine = engine
        self.sdk_version = sdk_version
        self.ui_type = ui_type

    def __str__(self):
        return "Engine = {0} {1},WeTest SDK = {2},UI={3}".format(self.engine, self.engine_version, self.sdk_version,
                                                                 self.ui_type)


class ElementBound(object):
    """
    Attributes:
        Element在屏幕上显示的位置和大小，(x,y)为中心点坐标系以屏幕左上角为坐标原点
        __________ x
        |
        |
        |
        y
        x:element与屏幕左侧的距离
        y:element与屏幕上边的距离
        width:element的宽
        height:element的高
        visible:是否可视化，在3D物体中
    """

    def __init__(self, x, y, width, height, visible=True):
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.visible = visible

    def __str__(self):
        return "  point = ({0},{1}) width = {2}  height = {3},visible={4}".format(self.x, self.y, self.width,
                                                                                  self.height, self.visible)

    def __repr__(self):
        return self.__str__()

    def __nonzero__(self):
        return self.visible


class WorldBound(object):
    """
        represents an axis aligned bounding box.
    """

    def __init__(self, _id, _existed):
        self.id = _id
        self.existed = _existed

        # center_x,center_y,center_z,the center of the bounding box
        self.center_x = 0
        self.center_y = 0
        self.center_z = 0

        # the extents of the box,
        self.extents_x = 0
        self.extents_y = 0
        self.extents_z = 0

    def __str__(self):
        return "center = ({0},{1},{2}) extents = ({3},{4},{5})".format(self.center_x, self.center_y, self.center_z,
                                                                       self.extents_x, self.extents_y, self.extents_z)


@six.add_metaclass(abc.ABCMeta)
class GameEngine(object):
    """
        Only support Unity engine at this time
    """

    def __init__(self, address, port):
        self.address = address
        self.port = port
        self.sdk_version = None
        self.socket = SocketClient(self.address, self.port)

    def engine_type(self):
        pass

    def get_sdk_version(self):
        """ 获取引擎集成的SDK的版本信息

        获取SDK的版本信息，GameSDKVersionInfo
        :return:
            返回的信息包括引擎版本，引擎和SDK版本信息,class:VersionInfo
            example:
                {‘engine_version’：‘5.1.0bf’，
                ‘engine’：‘Unity’,
                'sdk_version':10}
        :rtype: VersionInfo
        :raise:
            WeTestRuntimeError
        """
        ret = self.socket.send_command(Commands.GET_VERSION)
        engine = ret.get("engine", None)
        sdk_version = ret.get("sdkVersion", None)
        engine_version = ret.get("engineVersion", None)
        ui_type = ret.get("sdkUIType", None)
        version = VersionInfo(engine_version, engine, sdk_version, ui_type)
        return version

    def jump_building(self, name):
        """
            通过jump_building跳转到对应building
        :param name:
            jump_building的参数
        :Usage:
            >>>import wpyscripts.manager as manager
            >>>engine=manager.get_engine()
            >>>button=engine.jump_building()
        :return:
            status:0
        :rtype: Element
        :raise:
        """
        ret = self.socket.send_command(Commands.JUMP_BUILDING, name)
        if ret:
            return 0
        else:
            return None

    #
    # def input_text_set(self, element, text):
    #     """
    #             设置坐标
    #             :param :element,text
    #
    #             :Usage:
    #                 >>>import wpyscripts.manager as manager
    #                 >>>engine=manager.get_engine()
    #                 >>>text=engine.input_text_set(element,text)
    #             :return:成功与否
    #             """
    #     ret = self.socket.send_command(Commands.SET_INPUT_TEXT_VC, {"element": element, "text": text})
    #     if ret:
    #         return ret
    #     else:
    #         return None

    def find_element(self, name):
        """
            通过GameObject.Find查找对应的GameObject
        :param name:
            GameObject.Find的参数
        :Usage:
            >>>engine=GameEngine("127.0.0.1",12345)
            >>>button=engine.find_element('/Canvas/Panel/Button')
        :return:
            a instance of Element if find the GameObject,else return  None
            example:
            {"object_name":"/Canvas/Panel/Button",
            "instance":4257741}
        :rtype: Element
        :raise:
        """
        ret = self.socket.send_command(Commands.FIND_ELEMENTS, [name])
        if ret:
            ret = ret[0]
            if ret["instance"] == -1:
                return None
            else:
                return Element(ret["name"], ret["instance"])
        else:
            return None

    @abc.abstractmethod
    def find_elements_path(self, path):
        pass

    @abc.abstractmethod
    def get_element_text(self, element):
        pass

    @abc.abstractmethod
    def get_element_bound(self, element):
        pass

    def _get_dump_tree(self):
        """
        获取dump tree
        :return: xml string
        """
        ret = self.socket.send_command(Commands.DUMP_TREE)
        return ret

    def get_scene(self):
        """
            获取当前界面的scene名称
        :Usage:
            >>>import wpyscripts.manager as manager
            >>>engine=manager.get_engine()
            >>>current_scene=engine.get_scene()
        :return:当前scene的名称
        :raise: WeTestRuntimeError
        """
        ret = self.socket.send_command(Commands.GET_CURRENT_SCENE)
        return ret








# -*- coding: UTF-8 -*-
"""
Tencent is pleased to support the open source community by making GAutomator available.
Copyright (C) 2016 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.

"""
__author__ = 'minhuaxu wukenaihesos@gmail.com'

class Element(object):
    """
        in unity game Element is just GameObject
    Attributes:
        object_name:GameObject的全路径
        instance:GameObject Instance,the instance id of gameobject is always guarnnteed to be unique
    """

    def __init__(self, object_name, instance):
        self.__object_name = object_name
        self.__instance = instance

    @property
    def object_name(self):
        return self.__object_name

    @property
    def instance(self):
        return self.__instance

    def __str__(self):
        return "GameObject {0} Instance = {1}".format(self.object_name, self.instance)

    def __eq__(self, element):
        return hasattr(element, 'instance') and self.__instance == element.instance

    def __ne__(self, element):
        return not self.__eq__(element)

    def __repr__(self):
        return '<{0.__module__}.{0.__name__} (object_name="{1}", instance="{2}")>'.format(type(self), self.object_name, self.instance)




import yaml
from ga2.common.loggerConfig import logger
from ga2.device.device import DeviceType
from ga2.automation.by import By


class UIConfig:
    DeviceTypeMapper = {DeviceType.DEVICE_ANDROID: "android", DeviceType.DEVICE_IOS: "ios"}
    defaultConfig = None

    def __init__(self, path):
        self.path = path
        self._version = None
        self._elements = None
        self.load(path)


    @staticmethod
    def get_default():
        return UIConfig.defaultConfig

    @staticmethod
    def set_default(uiconfig):
        UIConfig.defaultConfig = uiconfig

    def set_default(self):
        UIConfig.defaultConfig = self

    def load(self, path):
        data = yaml.safe_load(open(path))
        assert ("elements" in data)
        if "version" in data:
            self._version = data["version"]

        self._elements = data["elements"]


    def get_locator(self, id, device_type):
        device_type_str = UIConfig.DeviceTypeMapper[device_type]
        if id not in self._elements:
            logger.error("id {} is not defined in uiconfig ".format(id))
            return (None, None)
        locators = self._elements[id]
        if isinstance(locators, dict):
            if device_type_str not in locators:
                logger.error("id {}'s locator in {} is not defined in uiconfig ".format(id, device_type))
                return (None, None)
            locators = locators[device_type_str]
        if len(locators) == 0:
            logger.error("id {}'s locator is not defined in uiconfig ".format(id))
            return (None, None)
        if len(locators) == 0:
            logger.error("id {}'s locator is not defined in uiconfig ".format(id))
            return (None, None)
        if len(locators) > 1:
            logger.error("multiple id {}'s locators is  defined in uiconfig. choose the first one for now ".format(id))
            return (None, None)
        locator = locators[0]
        for key in locator.keys():
            method = By.__dict__.get(key)
            params = locator[key]
            return (method, params)
        return (None, None)






# -*- coding: UTF-8 -*-
"""
Tencent is pleased to support the open source community by making GAutomator available.
Copyright (C) 2016 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.

"""
__author__ = 'minhuaxu wukenaihesos@gmail.com'
import six
if six.PY2:
    from exceptions import RuntimeError


class WeTestRuntimeError(RuntimeError):
    """
        与引擎、平台设备交互过程中遇到不可忽略的错误时抛出
    """
    pass


class WeTestEnvironmentError(WeTestRuntimeError):
    """
        环境变量设置有问题
    """
    pass


class WeTestPlatormError(WeTestRuntimeError):
    """
        平台连接错误
    """
    pass


class NetError(WeTestRuntimeError):
    """
        网络相关错误，初始化失败、网络断开连接
    """
    pass


class WeTestInvaildArg(WeTestRuntimeError):
    """
        错误的输入接口
    """
    pass


class WeTestSDKError(WeTestRuntimeError):
    """
        SDK错误
    """
    pass


class WeTestSDKTimeOut(WeTestRuntimeError):
    """
        SDK超时错误
    """
    pass


class WeTestInvaildSdkVersion(WeTestRuntimeError):
    """
        当前wpyscripts不支持对应的SDK
    """


class SceneTagError(WeTestRuntimeError):
    """
        性能数据设置标签错误
    """


class LoginError(WeTestRuntimeError):
    """
        登陆出错，比如QQ或微信没有账号密码等
    """


class UIAutomatorError(WeTestRuntimeError):
    """
        调用UIAutomator时错误
    """


class UIAutomatorLoginError(WeTestRuntimeError):
    """
        登录时出错
    """



import logging,os
import six
if six.PY2:
    import sys
    reload(sys)
    sys.setdefaultencoding('utf8')

LOG_FILE = 'wetest.log'


def makesure_dir_exist(path):
    existed=os.path.exists(path)
    if existed:
        return True
    else:
        os.mkdir(path)

handler = logging.StreamHandler()
fmt = '%(asctime)s - %(levelname)s - %(filename)s:%(lineno)s - %(name)s - %(message)s'

formatter = logging.Formatter(fmt)
handler.setFormatter(formatter)

logger = logging.getLogger()
logger.addHandler(handler)
logger.setLevel(logging.DEBUG)

# use_ai=False

#mode=os.environ.get("RUNNER_MODE")

# log_dir=os.environ.get("UPLOADDIR")
# #file_path = os.path.split(os.path.realpath(__file__))[0]
# file_path = os.getcwd()
# if log_dir:
#     file_path=os.path.abspath(os.path.join(log_dir,"ga_log.log"))
# else:
#     log_name="ga_log_{0}.log".format(os.environ.get("ADB_SERIAL",os.environ.get("IOS_SERIAL","")))
#     file_path=os.path.abspath(os.path.join(file_path,"..","..",log_name))
# file_handler=logging.FileHandler(file_path)
# file_handler.setFormatter(formatter)
# logger.addHandler(file_handler)
#


# def get_ga_logger():
#     if not get_ga_logger.instance:
#         get_ga_logger.instance=logging.getLogger("gautomator")
#     return get_ga_logger.instance
#
#
# get_ga_logger.instance=None


# -*- coding:UTF-8 -*-
"""
基于opencv的图片比较。依赖于logger模块和cv2helper模块。
"""
import os
import cv2
import time
import logging
import numpy as np
from .utils import MatchResult,_getPosition,transparent,judge_SimpleColor
from . import cv2helper

_DEBUG_HANDLERS = {}
logger = logging.getLogger('cvmatcher')
logger.setLevel(logging.DEBUG)
_filter = logging.Filter(name='cvmatcher')


def addHandler(hdlr):
    global _DEBUG_HANDLERS, logger
    clsName = hdlr.__class__.__name__
    if clsName not in _DEBUG_HANDLERS:
        hdlr.addFilter(_filter)
        logger.addHandler(hdlr)
        _DEBUG_HANDLERS[clsName] = hdlr


def pHash(img):
    """get image pHash value"""

    #调整为灰色
    if len(img.shape) > 2:
        img = cv2.cvtColor(img,cv2.COLOR_BGR2GRAY)

    h, w = img.shape[:2]
    if h > 32 and w > 32:
        #调整图片为32x32灰度图片
        img=cv2.resize(img,(32,32),interpolation=cv2.INTER_CUBIC)

        #创建二维列表
        h, w = img.shape[:2]
        vis0 = np.zeros((h,w), np.float32)
    elif h > 16 and w > 16:
        #调整图片为32x32灰度图片
        img=cv2.resize(img,(16,16),interpolation=cv2.INTER_CUBIC)

        #创建二维列表
        h, w = img.shape[:2]
        vis0 = np.zeros((h,w), np.float32)
    else:
        #调整图片为32x32灰度图片
        img=cv2.resize(img,(8,8),interpolation=cv2.INTER_CUBIC)

        #创建二维列表
        h, w = img.shape[:2]
        vis0 = np.zeros((h,w), np.float32)

#     #调整图片为32x32灰度图片
#     img=cv2.resize(img,(32,32),interpolation=cv2.INTER_CUBIC)
#
#     #创建二维列表
#     h, w = img.shape[:2]
#     vis0 = np.zeros((h,w), np.float32)
    vis0[:h,:w] = img       #填充数据

    #二维Dct变换
    vis1 = cv2.dct(cv2.dct(vis0))
    #cv.SaveImage('a.jpg',cv.fromarray(vis0)) #保存图片
    vis1.resize(8,8)

    #把二维list变成一维list
    img_list = vis1.flatten()#flatten(vis1.tolist())

    #计算均值
    avg = sum(img_list)*1./len(img_list)
    avg_list = ['0' if i<avg else '1' for i in img_list]

    #得到哈希值
    return ''.join(['%x' % int(''.join(avg_list[x:x+4]),2) for x in range(0,64,4)])

def hamming(h1, h2):
    '''汉明距离'''
    cnt = len(h1)
    total = 0
    for i in range(cnt):
        c1 = int(h1[i],16)
        c2 = int(h2[i],16)
        h, d = 0, c1 ^ c2

        while d:
            h += 1
            d &= d - 1
        #print h
        total += h
    return total

def compareByHamming(img1, img2, similarity=5):
    if not os.path.exists(img1):
        logger.error('%s not found' % img1)
        return False
    if not os.path.exists(img2):
        logger.error('%s not found' % img2)
        return False
    imgData1 = cv2.imread(img1)
    imgData2 = cv2.imread(img2)
    h1 = pHash(imgData1)
    h2 = pHash(imgData2)
    dis = hamming(h1, h2)
    if dis <= similarity:
        return True
    return False

class Hist(object):
    def _getImageData(self, filePath, gray):
        if not os.path.exists(filePath):
            logger.error('file not found:%s' % filePath)
            return None
        imgData = cv2helper.imread(filePath)
        if imgData is None:
            logger.error('invalid image format!')
            return None

        channel = imgData.shape[2]
        if not gray or (channel == 1 and gray):
            return imgData

        imgData = cv2helper.cvtColor(imgData)
        if imgData is None:
            logger.error('failed to covert to gray image')
            return None
        return imgData

    def _getHist(self, grayImg, colorBin=8):
        hist = cv2.calcHist([grayImg], [0], None, [colorBin], [0, 255])
        return cv2.normalize(hist, None).flatten()  # opencv 3.x

    def compare(self, img1, img2, gray=1, method='CHISQR', **kwargs):
        if gray != 1:
            raise NotImplementedError('gray must be 1')

        if method != 'CHISQR':
            raise NotImplementedError('method must be CHISQR')

        img1Data = None
        if isinstance(img1, str):
            img1Data = self._getImageData(img1, gray)

        elif cv2helper.isNdarray(img1):
            if gray and len(img1.shape) == 3:
                img1Data = cv2helper.cvtColor(img1)
            else:
                img1Data = img1
        else:
            raise TypeError('unknown type:%s' % type(img1))

        img2Data = None
        if isinstance(img2, str):
            img2Data = self._getImageData(img2, gray)
        elif cv2helper.isNdarray(img2):
            if gray and len(img2.shape) == 3:
                img2Data = cv2helper.cvtColor(img2)
            else:
                img2Data = img2
        else:
            raise TypeError('unknown type:%s' % type(img2))

        if img1Data is None or img2Data is None:
            raise ValueError('empty data')

        colorBin = 8
        if 'colorBin' in kwargs:
            colorBin = kwargs['colorBin']
        hist1 = self._getHist(img1Data, colorBin=colorBin)
        hist2 = self._getHist(img2Data, colorBin=colorBin)

        # method - cv2.HISTCMP_CHISQR, cv2.HISTCMP_BHATTACHARYYA, cv2.HISTCMP_CORREL, cv2.HISTCMP_INTERSECT
        rlt = cv2.compareHist(hist1, hist2, cv2.HISTCMP_CHISQR)  # [0, max), 0为最匹配， 取个经验值，小于0.1为相似
        #         print(cv2.compareHist(hist1, hist2, cv2.HISTCMP_BHATTACHARYYA))
        #         print(cv2.compareHist(hist1, hist2, cv2.HISTCMP_CORREL))
        #         print(cv2.compareHist(hist1, hist2, cv2.HISTCMP_INTERSECT))
        logger.info('hist result %s'%str(rlt))
        return rlt

    def similar(self, img1, img2, similarity=0.1, gray=1, method='CHISQR', **kwargs):
        rlt = self.compare(img1, img2, gray, method)
        return (rlt <= similarity)


class TemplateMatcher(object):
    _SCALE_RATIO_CACHE = {}
    _DEBUG_HANDLERS = {}

    def __init__(self):

        self._hist = Hist()

    #         if debug:
    #             selfpid = os.getpid()
    #             logDir = getEnv('WORKSPACE', os.getcwd())
    #             filePath=os.path.join(logDir, 'cvmatcher_{name}.log'.format(name=selfpid))
    #
    #             mode = 'w'
    #             if os.path.exists(filePath):
    #                 mode = 'a'
    #
    #             if selfpid not in TemplateMatcher._DEBUG_HANDLERS: # 避免重复
    #                 if int(getEnv('CVMATCHER_STDOUT', '0')):
    #                     _stdout = handler.DebugStreamHandler()
    #                     _stdout.addFilter(_filter)
    #                     logger.addHandler(_stdout)
    #
    #                 _hdlr = handler.DebugFileHandler(filePath, mode)
    #                 _hdlr.addFilter(_filter)
    #                 logger.addHandler(_hdlr)
    #                 TemplateMatcher._DEBUG_HANDLERS[selfpid] = _hdlr

    def _reset(self, trustSimilarity, **kwargs):
        self._maxHistSimilarity = 0
        self._minHistSimilarity = 1000
        self._maxSimilarity = 0
        self._minSimilarity = 1000
        self._matchCount = 0
        self._similarity = 0
        self._histSimilarity = 0

        self._trustSimilarity = trustSimilarity
        self._badTryCount = kwargs.get('badTryCount', 2)
        self._zoom = kwargs.get('zoom', 0)
        self._zoomOutDelta = kwargs.get('zoomOutDelta', 0.2)
        self._zoomInDelta = kwargs.get('zoomInDelta', 0.2)
        self._experienceSimilarity = kwargs.get('experienceSimilarity', 0.65)
        if self._experienceSimilarity >= trustSimilarity - 0.05:
            self._experienceSimilarity = trustSimilarity - 0.05
        self._lessGoodSimilarity2 = kwargs.get('lessGoodSimilarity2', 0.6)
        if self._lessGoodSimilarity2 >= trustSimilarity - 0.1:
            self._lessGoodSimilarity2 = trustSimilarity - 0.1
        self._untrustSimilarity = kwargs.get('untrustSimilarity', 0.3)
        self._histTrustSimilarity = kwargs.get('histTrustSimilarity', 0.1)
        self._histUntrustSimilarity = kwargs.get('histUntrustSimilarity', 10)
        self._use_histSimilarity = True

    def _getImageData(self, filePath, gray):
        if not os.path.exists(filePath):
            logger.error('file not found:%s' % filePath)
            return None
        imgData = cv2helper.imread(filePath)
        if imgData is None:
            logger.error('invalid image format!')
            return None

        channel = imgData.shape[2]
        if not gray or (channel == 1 and gray):
            return imgData

        imgData = cv2helper.cvtColor(imgData)
        if imgData is None:
            logger.error('failed to covert to gray image')
            return None
        return imgData

    def _getMultiResult(self, target, loc, distance=5):
        """获取多个相似对象"""
        if loc[0].size < 2:
            return [target]
        st = time.time()
        pts = zip(*loc[::-1])
        tmp = [target]
        for pt in pts:
            if abs(pt[0] - target[0]) > distance or abs(pt[1] - target[1]) > distance:
                tmp.append(pt)

        rlt = []
        for x in range(2):
            tmp.sort(key=lambda pt: pt[x])  # 坐标点(x,y)，先x排序，后y排序
            rlt = tmp[:]
            i = 0
            while i < len(tmp) - 1:
                pt_i = tmp[i]
                j = i + 1
                while j < len(tmp):
                    pt_j = tmp[j]
                    if abs(pt_j[0] - pt_i[0]) < distance and abs(pt_j[1] - pt_i[1]) < distance:
                        if pt_j in rlt:
                            if pt_j == target and pt_i in rlt:
                                rlt.remove(pt_i)
                            else:
                                rlt.remove(pt_j)
                    else:
                        break
                    j = j + 1
                i = j
            tmp = rlt

        et = time.time()
        logger.info('time cost:%s' % str(et - st))

        return rlt

    def _matchTemplateWithSeperateChannels(self, tplData, scnData, rect):
        if len(tplData.shape) < 3:
            # 已经是单通道，不再需要分通道
            return True

        tplImgHeight, tplImgWidth = tplData.shape[:2]
        newTplImg = tplData.copy()

        distance = min(tplImgHeight, tplImgWidth)
        t = cv2.split(newTplImg)
        s = cv2.split(scnData)
        max_similarity = 0
        max_rc = None
        c = 0
        for i in range(3):
            rlt = cv2helper.matchTemplate(s[i], t[i])
            _, max_val, _, max_loc = cv2helper.minMaxLoc(rlt)
            if max_val > max_similarity:
                max_similarity = max_val
                max_rc = max_loc
                c = i
        logger.debug('rc:%s' % str(rect))
        logger.debug('channel:%s' % c)
        logger.debug('channel_rc:%s' % str(max_rc))

        if abs(max_rc[0] - rect[0]) < distance and abs(max_rc[1] - rect[1]) < distance:
            return True
        return False

    def _validate(self, img1, img2, pt, print_log=False):
        img1 = img1.copy()
        h, w = img1.shape[:2]
        img2 = cv2helper.roi(img2.copy(), (pt[0], pt[1], pt[0] + w, pt[1] + h))

        # 直方图256 bins验证， 在遇到23_t_1时判断失败
        histValue = Hist().compare(img1, img2, colorBin=256)  # 10 for 256, 0.1 for 8 (color bin)
        if print_log:
            #             print('==============>hist:  %s' % (histValue))
            logger.debug('----------pt:%s--' % str(pt))
            logger.debug('----------hist check--:%s--' % histValue)

        if histValue > 50:
            return 2
        elif histValue <= 10:
            return 0
        else:
            return 1

        # phash验证

    #         p1 = pHash(img1)
    #         p2 = pHash(img2)
    #         pHM = hamming(p1, p2)
    #         if print_log:
    # #             print('============>hamming distance:  %s' % (pHM))
    #             logger.debug('---------->pt:%s--phash and hamming distance check--:%s<---------' % (str(pt), pHM))
    #
    #         if pHM > 15:
    #             return 2
    #
    #         if histValue <= 10 and pHM <= 15:
    #             return 0
    #
    #         return 1
    #         if rlt < 10:
    #             return 0
    #         elif rlt >= 10 and rlt < 50:
    #             return 1
    #         else:
    #             return 2

    def _matchTemplate(self, tplImg, scnImg, scale=1.0, similarity=0.7, index=-1):
        tplImgHeight, tplImgWidth = tplImg.shape[:2]
        scnImgHeight, scnImgWidth = scnImg.shape[:2]
        newTplImg = tplImg.copy()
        if scale != 1.0:
            tplImgHeight = int(tplImgHeight * scale)
            tplImgWidth = int(tplImgWidth * scale)
            newTplImg = cv2helper.resize(tplImg, tplImgWidth, tplImgHeight)

        if tplImgHeight > scnImgHeight or tplImgWidth > scnImgWidth:
            return (0, None)

        self._matchCount = self._matchCount + 1
        rlt = cv2helper.matchTemplate(scnImg.copy(), newTplImg.copy())
        _, max_val, _, max_loc = cv2helper.minMaxLoc(rlt)
        logger.debug('scale:%s' % scale)
        logger.debug('height:%s' % tplImgHeight)
        logger.debug('width:%s' % tplImgWidth)
        logger.debug('searched similarity:%s' % max_val)
        logger.debug('searched point:%s' % str(max_loc))

        self._similarity = max_val
        if max_val > self._maxSimilarity:
            self._maxSimilarity = max_val
        if max_val < self._minSimilarity:
            self._minSimilarity = max_val

        if max_val > similarity:
            need_validation = True
            if isinstance(index, int) and index >= 0:
                loc = cv2helper.npwhere(rlt >= similarity)
                cnt = loc[0].size

                if cnt > 1:
                    distance = min(tplImgHeight, tplImgWidth)
                    logger.debug('found count:%s' % cnt)
                    logger.debug('distance:%s' % distance)

                    tmpPoints = self._getMultiResult(max_loc, loc, distance)
                    logger.debug('distance:%s' % distance)

                    # todo: check whether it is real suitable
                    pts = []
                    for pt in tmpPoints:
                        #                         # phash 验证
                        #                         p1 = pHash(newTplImg.copy())
                        #                         roiImg = cv2helper.roi(scnImg.copy(), (pt[0], pt[1], pt[0]+tplImgWidth, pt[1]+tplImgHeight))
                        #                         p2 = pHash(roiImg)
                        #                         pHM = hamming(p1, p2)
                        #                         logger.debug('pt:%s--hamming distance--:%s' % (str(pt),pHM))
                        #                         # 三通道模板匹配验证
                        #                         matched = self._matchTemplateWithSeperateChannels(newTplImg.copy(), scnImg.copy(), (pt[0],pt[1],0,0))
                        #                         logger.debug('pt:%s--hist check--:%s' % (str(pt),matched))
                        #                         if matched:
                        #                             pts.append(pt)
                        # 直方图验证
                        validation_rlt = self._validate(newTplImg, scnImg, pt)
                        if validation_rlt == 0:
                            pts.append(pt)

                    finalCount = len(pts)
                    if finalCount > index:
                        tmp = pts[index]
                        if tmp == max_loc:
                            need_validation = False
                        max_loc = tmp
                    logger.info('final count:%s' % finalCount)
                    logger.info('index:%s' % index)
                    logger.info('point:%s' % str(max_loc))

            # todo: 找到后需要再次通过直方图确认，如果相似度很高，则不必再验证
            if need_validation and max_val < 0.8:
                if self._validate(newTplImg, scnImg, max_loc, True) > 1:
                    # 很大可能不相似
                    logger.info('not similarity after validation')
                    max_loc = None
        #                     max_val = 0

        rect = None
        if max_loc:
            rect = (max_loc[0], max_loc[1], tplImgWidth, tplImgHeight)
        return (max_val, rect)

    def _roi(self, tpl, img, rc, scaleRatioDelta):
        th, tw = tpl.shape[:2]
        ih, iw = img.shape[:2]
        l, t, w, h = rc
        delta = int(max(th * scaleRatioDelta, tw * scaleRatioDelta, 20)) + 1
        _l = l - delta
        _t = t - delta
        _r = l + w + delta
        _b = t + h + delta
        if _l < 0:
            _l = 0
        if _t < 0:
            _t = 0
        if _r > iw:
            _r = iw - 1
        if _b > ih:
            _b = ih - 1
        newImg = img[_t:_b, _l:_r].copy()
        return (newImg, (_l, _t))

    def _getHistSimilarity(self, tpl, scene, scale, rc):
        if not self._use_histSimilarity:
            return 1000

        tplImgHeight, tplImgWidth = tpl.shape[:2]
        newTplImg = tpl.copy()
        if scale != 1.0:
            tplImgHeight = int(tplImgHeight * scale)
            tplImgWidth = int(tplImgWidth * scale)
            newTplImg = cv2helper.resize(tpl, tplImgWidth, tplImgHeight)

        l, t, w, h = rc
        newSceneImg = scene[t:(t + h), l:(l + w)].copy()

        histSimilarity = self._hist.compare(newTplImg, newSceneImg)
        self._histSimilarity = histSimilarity
        if histSimilarity > self._maxHistSimilarity:
            self._maxHistSimilarity = histSimilarity
        if histSimilarity < self._minHistSimilarity:
            self._minHistSimilarity = histSimilarity
        logger.debug('hist similarity %s'%histSimilarity)
        return histSimilarity

    def _scale_direction(self, templ, image, direction, zoomMax, scale, lastMaxVal, rocPoint):

        logger.debug('scale direction, direction:%s' % direction)
        logger.debug('zoomMax:%s' % zoomMax)
        logger.debug('scale:%s' % scale)
        logger.debug('lastMaxValue:%s' % lastMaxVal)

        rocImage = image.copy()
        tryCount = 0
        while abs(1.0 - scale) < zoomMax:
            if lastMaxVal > self._experienceSimilarity:
                delta = 0.01
            elif lastMaxVal > self._lessGoodSimilarity2:
                delta = 0.02
            else:
                delta = 0.03
            scale = round(scale + direction * delta, 3)

            max_val, rect = self._matchTemplate(templ, rocImage, scale, self._trustSimilarity)
            if rect is None:
                break

            if max_val >= self._lessGoodSimilarity2:
                # 相似度必须大于某一个阈值，直方图比较才作为参考
                histSimilarity = self._getHistSimilarity(templ, image, scale, rect)
                if histSimilarity <= self._histTrustSimilarity:
                    if max_val >= self._experienceSimilarity:
                        # 如果相似度 > experienceSimilarity 并且 直方图比较结果 < histTrustSimilarity， 则认为很相似，直接返回
                        return scale, max_val, self._calcRect(rect, rocPoint)

                    if rocPoint is None:
                        rocImage, rocPoint = self._roi(templ, image, rect, zoomMax)

            time.sleep(0.05)  # 必须加，释放CPU资源
            if max_val < self._trustSimilarity:
                if max_val < lastMaxVal and max_val < self._experienceSimilarity:
                    # 比上次匹配结果更差，但是相似度小于experienceSimilarity
                    tryCount += 1
                    if tryCount > self._badTryCount:
                        break
                else:
                    tryCount = 1
                lastMaxVal = max_val
                continue

            return scale, max_val, self._calcRect(rect, rocPoint)

        return scale, None, None

    def _calcRect(self, rect, rocPoint):
        if rocPoint is None:
            rocPoint = (0, 0)
        l, t, w, h = rect
        x, y = rocPoint
        rect = (l + x, t + y, w, h)
        return rect

    def _scaleMatch(self, templ, image, similarity, maxValue, maxRect, **kwargs):
        lastMaxVal = maxValue

        scale = 1.0
        rocPoint = None
        rocImage = image.copy()
        if maxValue >= self._lessGoodSimilarity2:
            # 相似度必须大于某一个阈值，直方图比较才作为参考
            histSimilarity = self._getHistSimilarity(templ, image, scale, maxRect)
            if histSimilarity <= self._histTrustSimilarity:
                if maxValue >= self._experienceSimilarity:
                    # 如果相似度 > experienceSimilarity 并且 直方图比较结果 < histTrustSimilarity， 则认为很相似，直接返回
                    return scale, maxValue, maxRect

                rocImage, rocPoint = self._roi(templ, image, maxRect, self._zoomOutDelta)

        zoomAttr = {1: [0, 0, 0], -1: [0, 0, 0]}

        if self._zoom == 0:
            # 计算该先往哪方面缩放
            delta = 0.03
            if lastMaxVal > self._experienceSimilarity:
                delta = 0.01
            elif lastMaxVal > self._lessGoodSimilarity2:
                delta = 0.02

            zoom_out_scale = 1.0 + delta
            zoom_out_max_val, zoom_out_rect = self._matchTemplate(templ, rocImage, zoom_out_scale,
                                                                  self._trustSimilarity)
            if zoom_out_rect is None:
                return 1.0, None, None

            if zoom_out_max_val >= self._trustSimilarity:
                return zoom_out_scale, zoom_out_max_val, self._calcRect(zoom_out_rect, rocPoint)

            zoom_in_scale = 1.0 - delta
            zoom_in_max_val, zoom_in_rect = self._matchTemplate(templ, rocImage, zoom_in_scale, self._trustSimilarity)
            if zoom_in_rect is None:
                return 1.0, None, None

            if zoom_in_max_val >= self._trustSimilarity:
                return zoom_in_scale, zoom_out_max_val, self._calcRect(zoom_in_rect, rocPoint)

            zoomAttr[1] = [self._zoomOutDelta, zoom_out_scale, zoom_out_max_val]
            zoomAttr[-1] = [self._zoomInDelta, zoom_in_scale, zoom_in_max_val]
            if zoom_out_max_val >= zoom_in_max_val:
                direction = 1
            else:
                direction = -1

        else:
            direction = self._zoom
            scale = 1.0
            zoomAttr[direction][1] = 1.0
            zoomAttr[direction][2] = lastMaxVal
            if direction == 1:
                zoomAttr[direction][0] = self._zoomOutDelta
            else:
                zoomAttr[direction][0] = self._zoomInDelta

        logger.debug('direction:%s' %direction)
        logger.debug('zoom:%s' % self._zoom)
        logger.debug('zoom attrs:%s' % str(zoomAttr))

        scale, max_val, rect = self._scale_direction(templ, rocImage, direction, zoomAttr[direction][0],
                                                     zoomAttr[direction][1], zoomAttr[direction][2], rocPoint)

        logger.debug('direction:%s' % direction)
        logger.debug('scale:%s' % scale)
        logger.debug('max_val:%s' % max_val)
        logger.debug('rect:%s' % str(rect))

        if rect is not None:
            return scale, max_val, rect

        if self._zoom == 0:
            direction = direction * -1
            scale, max_val, rect = self._scale_direction(templ, rocImage, direction, zoomAttr[direction][0],
                                                         zoomAttr[direction][1], zoomAttr[direction][2], rocPoint)
            if rect is not None:
                return scale, max_val, rect

        return 1.0, None, None

    def _searchImage(self, tplData, sceneData, similarity, **kwargs):
        sceneImg = sceneData.copy()
        position = kwargs.get('position')
        logger.debug('templateImg position:%s' % str(position))
        is_transparent = kwargs.get('is_transparent')
        logger.debug('templateImg is_transparent:%s' % is_transparent)
        if position and len(position) == 4:
            logger.info('TemplateImg position is locked,crop Img!')
            crop_position = _getPosition(tplData, sceneData, position)
            crop_x_min, crop_y_min, crop_x_max, crop_y_max = crop_position
            sceneData = sceneData[crop_y_min:crop_y_max, crop_x_min:crop_x_max]
        else:
            logger.info('TemplateImg position is unlocked')
            crop_x_min, crop_y_min, crop_x_max, crop_y_max = 0, 0, 0, 0


        templ_h, templ_w = tplData.shape[:2]
        sceneImage_h, sceneImage_w = sceneData.shape[:2]

        if templ_h > sceneImage_h or templ_w > sceneImage_w:
            logger.error('tpl image size is smaller than screen' )
            return (None, None)

        res_simpleColor=judge_SimpleColor(tplData)
        scaleRatio = 1.0
        if is_transparent==1 or res_simpleColor:
            sim,rect=transparent(tplData,sceneData)
            if position and len(position) == 4:
                if sim>=0.6:
                    return ((rect[0]+crop_x_min, rect[1]+crop_y_min, rect[0] + rect[2]+crop_x_min, rect[1] + rect[3]+crop_y_min), scaleRatio)
            else:
                if sim>=0.8:
                    return ((rect[0] + crop_x_min, rect[1] + crop_y_min, rect[0] + rect[2] + crop_x_min,
                             rect[1] + rect[3] + crop_y_min), scaleRatio)
        if 'scaleRatio' in kwargs:
            scaleRatio = kwargs.pop('scaleRatio')
            max_val, rect = self._matchTemplate(tplData, sceneData, scaleRatio, similarity)
        else:
            max_val, rect = self._matchTemplate(tplData, sceneData, 1.0, similarity)

        if rect is None:
            return (None, None)

        if max_val < similarity:

            if 'scaleMatch' in kwargs and kwargs['scaleMatch']:
                scaleRatio, max_val, rect = self._scaleMatch(tplData, sceneData, similarity, max_val, rect, **kwargs)
                if rect:
                    x, y, w, h = rect
                    return ((x+crop_x_min, y+crop_y_min, x + w+crop_x_min, y + h+crop_y_min), scaleRatio)
            return (None, None)

        return ((rect[0]+crop_x_min, rect[1]+crop_y_min, rect[0] + rect[2]+crop_x_min, rect[1] + rect[3]+crop_y_min), scaleRatio)


    def searchImage(self, templateImage, sceneImage, similarity=0.7, gray=False, **kwargs):
        result = MatchResult()
        result.method = 'template'
        st = time.time()
        title = ''
        self._reset(similarity, **kwargs)
        tplData = None
        if isinstance(templateImage, str):
            title = os.path.splitext(os.path.split(templateImage)[-1])[0]
            tplData = self._getImageData(templateImage, gray)
        elif cv2helper.isNdarray(templateImage):
            if gray and len(tplData.shape) == 3:
                tplData = cv2helper.cvtColor(templateImage)
            else:
                tplData = templateImage
        else:
            raise TypeError('unknown type:%s' % type(templateImage))


        sceneData = None
        if isinstance(sceneImage, str):
            title = '%s_%s' % (title, os.path.splitext(os.path.split(sceneImage)[-1])[0])
            sceneData = self._getImageData(sceneImage, gray)
        elif cv2helper.isNdarray(sceneImage):
            if gray and len(sceneImage.shape) == 3:
                sceneData = cv2helper.cvtColor(sceneImage)
            else:
                sceneData = sceneImage
        else:
            raise TypeError('unknown type:%s' % type(sceneImage))

        if tplData is None or sceneData is None:
            et = time.time()
            time_cost = et - st
            logger.debug('failed to get image data')
            result.time=time_cost
            result.parameters = kwargs
            return result

        if isinstance(templateImage, str) and int(kwargs.get('cache', '1')) and \
                templateImage in TemplateMatcher._SCALE_RATIO_CACHE:
            kwargs['scaleRatio'] = TemplateMatcher._SCALE_RATIO_CACHE[templateImage]

        rect, scaleRatio = self._searchImage(tplData, sceneData, similarity, **kwargs)
        if rect:
            index = kwargs.get('index', -1)
            if index >= 0:
                logger.debug('index:%s' % index)
                finalSimilarity = self._similarity
                if finalSimilarity > similarity:
                    # 取标准相似度值，否则结果只有1个
                    finalSimilarity = similarity

                _, _rect = self._matchTemplate(tplData, sceneData, scaleRatio, finalSimilarity, index)
                if rect:
                    l, t, w, h = _rect
                    rect = (l, t, l + w, t + h)

            if rect:
                if isinstance(templateImage, str) and int(kwargs.get('cache', '1')) and \
                        scaleRatio != 1.0 and templateImage not in TemplateMatcher._SCALE_RATIO_CACHE:
                    TemplateMatcher._SCALE_RATIO_CACHE[templateImage] = scaleRatio

                if isinstance(templateImage, str):
                    et = time.time()
                    time_cost = et - st
                    logger.debug('found, matched count:%s' % self._matchCount)
                    logger.debug('rect: %s' % str(rect))
                    logger.debug('final scaleRatio:%s' % scaleRatio)
                    logger.debug('max hist:%s' % self._maxHistSimilarity)
                    logger.debug('min similarity:%s' % self._minSimilarity)
                    logger.debug('histSimilarity:%s' % self._histSimilarity)
                    logger.debug('similarity:%s' % self._similarity)
                    logger.debug('time cost:%s' % time_cost)
                    logger.debug('matched count:%s' % self._matchCount)

                if 'showResult' in kwargs and kwargs['showResult']:
                    self._showResult(tplData, sceneData, [rect], "OK_%s" % title)


                result.rect = rect
                et = time.time()
                time_cost = et - st
                result.time = time_cost
                result.scale_ratio = scaleRatio
                result.parameters = kwargs
                return result

        if isinstance(templateImage, str):
            et = time.time()
            time_cost = et - st
            result.time = time_cost
            result.parameters = kwargs
            logger.debug('none, matched count:%s' % self._matchCount)
            logger.debug('rect: %s' % str(rect))
            logger.debug('max hist:%s' % self._maxHistSimilarity)
            logger.debug('min similarity:%s' % self._minSimilarity)
            logger.debug('time cost:%s' % time_cost)
            logger.debug('matched count:%s' % self._matchCount)
        return result

    def _showResult(self, templateData, sceneData, rects, title='show'):
        for rect in rects:
            l, t, r, b = rect
            cv2helper.eval2('cv2.rectangle')(sceneData, (l, t), (r, b), (0, 0, 255), 2)

        h1, w1 = templateData.shape[:2]
        h2, w2 = sceneData.shape[:2]

        if len(sceneData.shape) > 2:
            vis = cv2helper.eval2('np.zeros')((max(h1, h2), w1 + w2, 3))
        else:
            vis = cv2helper.eval2('np.zeros')((max(h1, h2), w1 + w2))

        vis[:h1, :w1] = templateData
        vis[:h2, w1:w1 + w2] = sceneData

        fshow = '%s.jpg' % title
        cv2.imwrite(fshow, vis)
        tmp = cv2.imread(fshow)
        cv2helper.show(tmp, title)
        os.remove(fshow)


#-*- coding: UTF-8 -*-
__author__ = 'minhuaxu'

# Licensed to the Software Freedom Conservancy (SFC) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The SFC licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging, time
import socket
import string
import base64

try:
    import http.client as httplib
    from urllib import request as url_request
    from urllib import parse
except ImportError: # above is available in py3+, below is py2.7
    import httplib as httplib
    import urllib2 as url_request
    import urlparse as parse

from ga2.cloud.httptools import utils
from .errorhandler import ErrorCode,ErrorHandler


LOGGER = logging.getLogger(__name__)


class Method(object):
    GET="GET"
    POST="POST"
    PUT="PUT"

class Request(url_request.Request):
    """
    Extends the url_request.Request to support all HTTP request types.
    """

    def __init__(self, url, data=None, method=None):
        """
        Initialise a new HTTP request.

        :Args:
        - url - String for the URL to send the request to.
        - data - Data to send with the request.
        """
        if method is None:
            method = data is not None and 'POST' or 'GET'
        elif method != 'POST' and method != 'PUT':
            data = None
        self._method = method
        url_request.Request.__init__(self, url, data=data)

    def get_method(self):
        """
        Returns the HTTP method used by this request.
        """
        return self._method


class Response(object):
    """
    Represents an HTTP response.
    """

    def __init__(self, fp, code, headers, url):
        """
        Initialise a new Response.

        :Args:
        - fp - The response body file object.
        - code - The HTTP status code returned by the server.
        - headers - A dictionary of headers returned by the server.
        - url - URL of the retrieved resource represented by this Response.
        """
        self.fp = fp
        self.read = fp.read
        self.code = code
        self.headers = headers
        self.url = url

    def close(self):
        """
        Close the response body file object.
        """
        self.read = None
        self.fp = None

    def info(self):
        """
        Returns the response headers.
        """
        return self.headers

    def geturl(self):
        """
        Returns the URL for the resource returned in this response.
        """
        return self.url


class HttpErrorHandler(url_request.HTTPDefaultErrorHandler):
    """
    A custom HTTP error handler.

    Used to return Response objects instead of raising an HTTPError exception.
    """

    def http_error_default(self, req, fp, code, msg, headers):
        """
        Default HTTP error handler.

        :Args:
        - req - The original Request object.
        - fp - The response body file object.
        - code - The HTTP status code returned by the server.
        - msg - The HTTP status message returned by the server.
        - headers - The response headers.

        :Returns:
        A new Response object.
        """
        return Response(fp, code, headers, req.get_full_url())


class RemoteConnection(object):
    """A connection with the Remote WebDriver server.

    Communicates with the server using the WebDriver wire protocol:
    https://github.com/SeleniumHQ/selenium/wiki/JsonWireProtocol"""

    _timeout = socket._GLOBAL_DEFAULT_TIMEOUT

    def __init__(self, remote_server_addr, keep_alive=False,timeout=socket._GLOBAL_DEFAULT_TIMEOUT):
        # Attempt to resolve the hostname and get an IP address.
        self.keep_alive = keep_alive
        self.error_handler=ErrorHandler()
        self._timeout=timeout

        parsed_url = parse.urlparse(remote_server_addr)
        addr = ""
        if parsed_url.hostname:
            try:
                netloc = socket.gethostbyname(parsed_url.hostname)
                addr = netloc
                if parsed_url.port:
                    netloc += ':%d' % parsed_url.port
                if parsed_url.username:
                    auth = parsed_url.username
                    if parsed_url.password:
                        auth += ':%s' % parsed_url.password
                    netloc = '%s@%s' % (auth, netloc)
                remote_server_addr = parse.urlunparse(
                    (parsed_url.scheme, netloc, parsed_url.path,
                     parsed_url.params, parsed_url.query, parsed_url.fragment))
            except socket.gaierror:
                LOGGER.info('Could not get IP address for host: %s' % parsed_url.hostname)

        self._url = remote_server_addr
        if keep_alive:
            self._conn = httplib.HTTPConnection(
                str(addr), str(parsed_url.port), timeout=self._timeout)


    def _wrap_value(self,value):
        if isinstance(value, dict):
            converted = {}
            for key, val in value.items():
                converted[key] = self._wrap_value(val)
            return converted
        elif isinstance(value, list):
            return list(self._wrap_value(item) for item in value)
        else:
            return value


    def _check_response(self,response):
        if response:
            self.error_handler.check_response(response)
            response['value'] = self._unwrap_value(
                response.get('value', None))
            return response
        # If the server doesn't send a response, assume the command was
        # a success
        return {'success': 0, 'value': None, 'sessionId': self.session_id}

    def _execute(self,path,method,params=None):
        """
        Send a command to the remote server.

        Any path subtitutions required for the URL mapped to the command should be
        included in the command parameters.

        :Args:
         - command - A string specifying the command to execute.
         - params - A dictionary of named parameters to send with the command as
           its JSON payload.
        """
        assert path is not None and method is not None, 'Url Method not None'
        data = utils.dump_json(params)
        path = string.Template(path).substitute(params)
        url = '{0}/{1}'.format(self._url, path)
        return self._request(method, url, body=data)


    def get(self,path,params=None):
        """
            Get method
        :param path:
        :param params:
        :return:
        """
        reponse=self._execute(path,Method.GET,params)
        if reponse:
            self.error_handler.check_response(reponse)
            return reponse
        else:
            raise RuntimeError("get url error")

    def post(self,path,params):
        """
            Post method
        :param path:
        :param params:
        :return:
        """
        reponse=self._execute(path,Method.POST,params)
        if reponse:
            self.error_handler.check_response(reponse)
            return reponse
        return True

    def _request(self, method, url, body=None):
        """
        Send an HTTP request to the remote server.

        :Args:
         - method - A string for the HTTP method to send the request with.
         - url - A string for the URL to send the request to.
         - body - A string for request body. Ignored unless method is POST or PUT.

        :Returns:
          A dictionary with the server's parsed JSON response.
        """
        LOGGER.debug('%s %s %s' % (method, url, body))

        parsed_url = parse.urlparse(url)

        if self.keep_alive:
            headers = {"Connection": 'keep-alive', method: parsed_url.path,
                       "User-Agent": "Python http auth",
                       "Content-type": "application/json;charset=\"UTF-8\"",
                       "Accept": "application/json"}
            if parsed_url.username:
                auth = base64.standard_b64encode('%s:%s' %
                       (parsed_url.username, parsed_url.password)).replace('\n', '')
                headers["Authorization"] = "Basic %s" % auth
            if body and method != 'POST' and method != 'PUT':
                body = None
            try:
                self._conn.request(method, parsed_url.path, body, headers)
                resp = self._conn.getresponse()
            except (httplib.HTTPException, socket.error):
                self._conn.close()
                raise

            statuscode = resp.status
        else:
            password_manager = None
            if parsed_url.username:
                netloc = parsed_url.hostname
                if parsed_url.port:
                    netloc += ":%s" % parsed_url.port
                cleaned_url = parse.urlunparse((parsed_url.scheme,
                                                   netloc,
                                                   parsed_url.path,
                                                   parsed_url.params,
                                                   parsed_url.query,
                                                   parsed_url.fragment))
                password_manager = url_request.HTTPPasswordMgrWithDefaultRealm()
                password_manager.add_password(None,
                                              "%s://%s" % (parsed_url.scheme, netloc),
                                              parsed_url.username,
                                              parsed_url.password)
                request = Request(cleaned_url, data=body.encode('utf-8'), method=method)
            else:
                request = Request(url, data=body.encode('utf-8'), method=method)

            request.add_header('Accept', 'application/json')
            request.add_header('Content-Type', 'application/json;charset=UTF-8')

            if password_manager:
                opener = url_request.build_opener(url_request.HTTPRedirectHandler(),
                                                  HttpErrorHandler(),
                                                  url_request.HTTPBasicAuthHandler(password_manager))
            else:
                opener = url_request.build_opener(url_request.HTTPRedirectHandler(),
                                                  HttpErrorHandler())
            resp = opener.open(request, timeout=self._timeout)
            statuscode = resp.code
            if not hasattr(resp, 'getheader'):
                if hasattr(resp.headers, 'getheader'):
                    resp.getheader = lambda x: resp.headers.getheader(x)
                elif hasattr(resp.headers, 'get'):
                    resp.getheader = lambda x: resp.headers.get(x)

        data = resp.read()
        try:
            if 300 <= statuscode < 304:
                return self._request('GET', resp.getheader('location'))
            body = data.decode('utf-8').replace('\x00', '').strip()
            if 399 < statuscode <= 510:
                return {'status': statuscode, 'value': body}
            content_type = []
            if resp.getheader('Content-Type') is not None:
                content_type = resp.getheader('Content-Type').split(';')
            if not any([x.startswith('image/png') for x in content_type]):
                try:
                    data = utils.load_json(body.strip())
                except ValueError:
                    if 199 < statuscode < 300:
                        status = ErrorCode.SUCCESS
                    else:
                        status = ErrorCode.UNKNOWN_ERROR
                    return {'status': status, 'value': body.strip()}

                assert type(data) is dict, (
                    'Invalid server response body: %s' % body)
                # Some of the drivers incorrectly return a response
                # with no 'value' field when they should return null.
                # if 'value' not in data:
                #     data['value'] = None
                return data
            else:
                data = {'status': 0, 'value': body.strip()}
                return data
        finally:
            LOGGER.debug("Finished Request")
            resp.close()


if __name__ == '__main__':
    remote=RemoteConnection("http://127.0.0.1:5000",keep_alive=True)
    data= remote.get("tasks")
    print(data)

    post_data={
        'id': 3,
        'title': u'minhuaxu',
        'description': u'Need to find a good Python tutorial on the web',
        'done': False
    }

    data=remote.post("tasks",post_data)
    print(data)



# -*- coding: UTF-8 -*-
import re
from ga2.device.device import Device
from ga2.device.device import DeviceType
from ga2.device.deviceOrientation import *
from ga2.device.iOS.iMobiledevice import IMobileDevice
from ga2.device.iOS.defaultScreenShoter import DefaultScreenshoter
from ga2.device.iOS.wdaManager import WdaManager
from ga2.device.iOS.iProxyManager import IProxyManager
from ga2.common.loggerConfig import logger
from ga2.common.cmdExecuter import CmdExecuter
from ga2.common.utils import *
import atexit


class IOSDevice(Device):
    '''
    param: wdaport: the local port you have forward to the device
    '''

    def __init__(self, udid,  gaport=None, screenshoter=None):
        if udid is None or udid is "":
            dev_list = IMobileDevice.get_device_list()
            assert (len(dev_list) > 0)
            udid = dev_list[0].decode("utf-8")
        self.imobile_cmder = IMobileDevice(udid)
        self.__iproxy = IProxyManager(udid)
        self._engine_connector = None
        self.__wda_connector = None
        self._local_engine_port = gaport

        def __get_wdaport_by_udid(udid):
            outputlines = CmdExecuter.execute_and_wait("ps -ef|grep \"iproxy [0-9]\+ 8100 " + udid + "\"")
            for line in outputlines:
                ret = re.search("iproxy ([0-9]+) 8100 " + udid, line.decode("utf-8"))
                if ret:
                    return ret.group(1)
            outputlines = CmdExecuter.execute_and_wait("ps -ef|grep \"iproxy [0-9]\+ 8100" + "\"")
            for line in outputlines:
                ret = re.search("iproxy ([0-9]+) 8100$", line.decode("utf-8"))
                if ret:
                    return ret.group(1)
            logger.warn("No wda port detected for device.. " + udid)
            return None

        wdaport = os.environ.get("WDA_LOCAL_PORT", __get_wdaport_by_udid(udid))
        if wdaport:
            logger.info("wda port detected: " + wdaport)
            self.__wda_connector = WdaManager(wdaport, os.environ.get("PLATFORM_IP", "127.0.0.1"))
        else:
            logger.warn("wda port is not set!!")
        self._screenshoter = screenshoter or DefaultScreenshoter(self.__wda_connector)
        atexit.register(self.__cleanup)
        pass

    def wda_session(self):
        if self.__wda_connector == None:
            logger.error("wda connector is not inited... plsease make sure wda is connectable")
        return self.__wda_connector.get_session()

    def __cleanup(self):
        if self.__iproxy:
            self.__iproxy.remove_all_forwards()


    '''
    params: appfile:path to .ipa file
    '''

    # def install(self,appfile):
    #     self.imobile_cmder.install(appfile)
    #
    # def uninstall(self,appid):
    #     self.imobile_cmder.uninstall(appid)

    def device_type(self):
        return DeviceType.DEVICE_IOS


    @callLog
    def launch_app(self, appid, activity=None, timeout=60, alert_handler=None):
        if self.__wda_connector is None:
            logger.error("wda connector is not inited... plsease make sure wda is connectable")
            return ErrType.ERR_WDA_NOT_RUNNING
        session = self.__wda_connector.new_session(bundleid=appid)
        alert_handler = alert_handler or self.__default_alert_callback
        start_time = time.time()
        for i in range(0, int(timeout/3)):
            cur_top_app = self.get_top_app()
            logger.debug("current top app is :" + cur_top_app + "target is: " + appid)
            if cur_top_app != appid:
                alert_handler(session)
            else:
                break
            time.sleep(3)
            if time.time() - start_time > timeout:
                logger.warn("launch_app timeout...")
                return ErrType.ERR_TIMEOUT
        return ErrType.ERR_SUCCEED

    def start_alert_handler(self, handler=None):
        if self.__wda_connector is None:
            logger.error("wda connector is not inited... plsease make sure wda is connectable")
            return ErrType.ERR_WDA_NOT_RUNNING
        handler = handler or self.__default_alert_callback
        logger.info("setting alert handler...")
        self.__wda_connector.get_session().set_alert_callback(handler)
        return ErrType.ERR_SUCCEED

    def __default_alert_callback(self, session):
        btns = set([u'稍后', u'稍后提醒我', u'不再提醒', u'无线局域网与蜂窝移动网络', u'允许', u'知道了', u'确定', u'好']).intersection(
            session.alert.buttons())
        if len(btns) == 0:
            logger.warn("Alert  not handled, buttons: " + ', '.join(session.alert.buttons()))
            return
        logger.info('alert handled：' + str(list(btns)[0]))
        session.alert.click(list(btns)[0])
        pass

    @callLog
    def get_top_app(self):
        if self.__wda_connector is None:
            logger.error("wda connector is not inited... plsease make sure wda is connectable")
            return None
        return self.__wda_connector.get_top_bundleid()

    # @callLog
    # def foreground(self,appid):
    #     self.wda_connector.get_session(appid).activate(1)

    '''
    For now , just kill the app launched by device interface. It's better to save a map as {appid:session}
    '''

    @callLog
    def kill_app(self, appid=None):
        self.__wda_connector.close_session(appid)

    @callLog
    def home(self):
        self.__wda_connector.wda_client.home()

    def forward(self, local_port, remote_port):
        self.__iproxy.forward(localport=local_port, remoteport=remote_port)

    def remove_forward(self, local_port):
        self.__iproxy.remove_forward(localport=local_port)

    '''
    input text to device
    param : str
    '''

    def text(self, content):
        self.__wda_connector.get_session().send_keys(content)

    '''
    get the current device display size
    return: two values  (width,height)
    '''

    def display_size(self):
        return self.__wda_connector.get_session().window_size()

    '''
    get the current device orientation
    return: enum DeviceOrientation value
    '''

    def orientation(self):
        return WDA_ORIENTATION_MAP[self.__wda_connector.get_session().orientation]

    '''
    touch at the given position 
    param: x,y : pixel pos  or rate of the current device screen.
    '''

    def touch(self, x, y):
        if 0 < x < 1 and 0 < y < 1:
            display_size = self.display_size()
            x = x * display_size[0]
            y = y * display_size[1]
        logger.info("tap :" + str(x) + str(y))
        self.__wda_connector.get_session().tap(x, y)

    '''
    double touch at the given position 
    param: x,y : pixel pos  or rate of the current device screen.
    '''

    def double_touch(self, x, y):
        if 0 < x < 1 and 0 < y < 1:
            display_size = self.display_size()
            x = x * display_size[0]
            y = y * display_size[1]
        self.__wda_connector.get_session().double_tap(x, y)

    '''
    note:  the param "duration" indicates the touchdown and hold time rather than the moving time.
    param: sx,sy,dx,dy: start and end pixel pos  or rate of the current device screen.
           duration: start coordinate press duration in seconds
    '''

    def drag(self, sx, sy, dx, dy, duration=0.3):
        if 0 < sx < 1 and 0 < sy < 1 and 0 < dx < 1 and 0 < dy < 1:
            display_size = self.display_size()
            sx = sx * display_size[0]
            sy = sy * display_size[1]
            dx = dx * display_size[0]
            dy = dy * display_size[1]
        self.__wda_connector.get_session().swipe(sx, sy, dx, dy, duration)

    '''
    touch and hold for duration at the given position
    param : x,y : pixel pos  or rate of the current device screen.
            duration: holding seconds
    '''

    def long_press(self, x, y, duration=1.5):
        if 0 < x < 1 and 0 < y < 1:
            display_size = self.display_size()
            x *= display_size[0]
            y *= display_size[1]
        self.__wda_connector.get_session().tap_hold(x, y, duration)


    def push(self, local_path, remote_path):
        logger.error("ios push unimplemented yet")
        return ErrType.ERR_UNIMPLEMENTED

    def pull(self, remote_path, local_path):
        logger.error("ios pull unimplemented yet")
        return ErrType.ERR_UNIMPLEMENTED

    def clear_app(self):
        logger.error("ios clear_app unimplemented yet")
        return ErrType.ERR_UNIMPLEMENTED

    def keyevent(self):
        logger.error("ios keyevent unimplemented yet")
        return ErrType.ERR_UNIMPLEMENTED

    click = touch
    swipe = drag


# -*- coding: UTF-8 -*-
from ga2.engine.engine import *
from ga2.engine.uielement import UIElement
from ga2.common.utils import *

class UE4Engine(GameEngine):
    def __init__(self, address, port):
        super(UE4Engine, self).__init__(address, port)
        self._callback_socket = None
        self._callback_functions = {}
        self._callback_thread = None

    def engine_type(self):
        return "UE4"


    def find_elements_path(self, path):
        logger.error("find_elements_path is not implemented int UE4")
        return ErrType.ERR_UNIMPLEMENTED

    def find_elements_by_component(self, name):
        logger.error("find_elements_by_component is not implemented int UE4")
        return ErrType.ERR_UNIMPLEMENTED

    def get_element_text(self, element):
        """
            获取UMG文字内容，支持UMultiLineEditableText，UTextBlockmUMultiLineEditableTextBox及其子类UMG类型
        :param element: 查找的Element

        :Usage:
            >>>element=engine.find_element('Button')
            >>>text=engine.get_element_text(element)
        :return:文字内容
        :raises WeTestInvaildArg,WeTestRuntimeError
        """
        if element is None:
            raise WeTestInvaildArg("Invalid Instance")
        ret = self.socket.send_command(Commands.GET_ELEMENT_TEXT, element.instance)
        return ret



    def get_element_bound(self, element):
        """
        获取GameObject在屏幕上的位置和长宽高
        :param element: 查找到的GameObject

        :Usage:
            >>>
            >>>element=engine.find_element('Button')
            >>>bound=engine.get_element_bound(element)
        :return:屏幕中的位置（x,y），左上角为原点，及长宽
            examples:
            {"x":0.5，
            "y":0.5.0,
            "width":0.2
            "height":0.1}
        :rtype: ElementBound
        :raises WeTestInvaildArg,WeTestRuntimeError
        """
        if element is None:
            raise WeTestInvaildArg("Invaild Instance,element is None")

        ret = self._get_elements_bound([element])
        print(ret)
        if ret:
            result = ret[0]
            if not result["existed"]:
                return None
            else:
                return ElementBound(result["x"], result["y"], result["width"], result["height"], result["visible"])
        return None

    def _get_elements_bound(self, elements):
        send_params = [e.object_name for e in elements]
        ret = self.socket.send_command(Commands.GET_ELEMENTS_BOUND, send_params)
        return ret




from ga2.common.loggerConfig import logger
from ga2.common.utils import callLog
from ga2.device.device import Device
from ga2.device.android.uiauto import uiautoManager
from ga2.device.android import androidDevice
from ga2.device.device import DeviceType
from ga2.common.utils import *
from ga2.cloud.reporter import Reporter
from ga2.common.WebDriverWait import WebDriverWait


class AutomationHelperNative:

    @staticmethod
    def __get_uiauto_elem_by_text(device, text):
        elem = device.uiauto_man.ui_device(text=text)
        if not elem.exists:
            elem = device.uiauto_man.ui_device(description=text)
        return elem if elem.exists else None

    @staticmethod
    def __get_ios_elem_by_text(device, text):
        session = device.wda_session()
        elem = session(name=text)
        if not elem.exists:
            elem = session(label=text)
            if not elem.exists:
                elem = session(value=text)
        return elem if elem.exists else None

    @staticmethod
    def __get_uiauto_elem_by_conditions(device, conditions):
        elem = device.uiauto_man.ui_device(**conditions)
        return elem if elem.exists else None

    @staticmethod
    def __get_ios_elem_by_conditions(device, conditions):
        session = device.wda_session()
        elem = session(**conditions)
        return elem if elem.exists else None

    @staticmethod
    def wait_element_by_conditions(device, conditions, timeout):
        if device:
            element = None
            find_method = None
            if device.device_type() == DeviceType.DEVICE_ANDROID:
                if device.uiauto_man == None:
                    logger.error("wait_element_by_text failed : uiauto_man is not set")
                    return None
                find_method = AutomationHelperNative.__get_uiauto_elem_by_conditions
            elif device.device_type() == DeviceType.DEVICE_IOS:
                session = device.wda_session()
                if session is None:
                    logger.error("wait_element_by_text failed : session is not set")
                    return None
                find_method = AutomationHelperNative.__get_ios_elem_by_conditions
            try:
                element = WebDriverWait(timeout, 2).until(find_method, device, conditions)
            except Exception as e:
                logger.warn("element wait timeout:" + str(conditions))
            return element
        return None

    @staticmethod
    def wait_element_by_text(device, text, timeout):
        if device:
            element = None
            find_method = None
            if device.device_type() == DeviceType.DEVICE_ANDROID:
                if device.uiauto_man == None:
                    logger.error("wait_element_by_text failed : uiauto_man is not set")
                    return None
                find_method = AutomationHelperNative.__get_uiauto_elem_by_text
            elif device.device_type() == DeviceType.DEVICE_IOS:
                session = device.wda_session()
                if session is None:
                    logger.error("wait_element_by_text failed : session is not set")
                    return None
                find_method = AutomationHelperNative.__get_ios_elem_by_text
            try:
                element = WebDriverWait(timeout, 2).until(find_method, device, text)
            except Exception as e:
                logger.warn("element wait timeout:" + text)
            return element
        return None

    @staticmethod
    def handle_element_by_text(device, text, operator, **kwargs):
        if not device:
            return None
        target_pos = None
        if device.device_type() == DeviceType.DEVICE_ANDROID:
            if device.uiauto_man == None:
                logger.error("handle_element_by_text failed : uiauto_man is not set")
                return None
            elem = AutomationHelperNative.__get_uiauto_elem_by_text(device, text)
            if elem:
                bound_info = elem.info[u'bounds']
                left, top, right, bottom = bound_info[u'left'], bound_info[u'top'], bound_info[u'right'], bound_info[
                    u'bottom']
                target_pos = ((left + right) / 2, (top + bottom) / 2)

        elif device.device_type() == DeviceType.DEVICE_IOS:
            session = device.wda_session()
            if session is None:
                logger.error("wait_element_by_text failed : session is not set")
                return None
            elem = AutomationHelperNative.__get_ios_elem_by_text(device, text)
            if elem:
                rect = elem.bound
                target_pos = (rect.x + rect.width / 2, rect.y + rect.height / 2)

        if target_pos:
            if is_cloud_mode():
                (width, height) = device.display_size()
                Reporter().screenshot_with_mark(width, height, target_pos[0], target_pos[1])

            operator(target_pos[0], target_pos[1], **kwargs)
        else:
            logger.warn("native elem not found : " + text)
        return target_pos

    @staticmethod
    def handle_element_by_conditions(device, conditions, operator, **kwargs):
        if not device:
            return None
        target_pos = None
        if device.device_type() == DeviceType.DEVICE_ANDROID:
            if device.uiauto_man == None:
                logger.error("handle_element_by_conditions failed : uiauto_man is not set")
                return None
            elem = AutomationHelperNative.__get_uiauto_elem_by_conditions(device, conditions)
            if elem:
                bound_info = elem.info[u'bounds']
                left, top, right, bottom = bound_info[u'left'], bound_info[u'top'], bound_info[u'right'], bound_info[
                    u'bottom']
                target_pos = ((left + right) / 2, (top + bottom) / 2)

        elif device.device_type() == DeviceType.DEVICE_IOS:
            session = device.wda_session()
            if session is None:
                logger.error("wait_element_by_text failed : session is not set")
                return None
            elem = AutomationHelperNative.__get_ios_elem_by_conditions(device, conditions)
            if elem:
                rect = elem.bound
                target_pos = (rect.x + rect.width / 2, rect.y + rect.height / 2)

        if target_pos:
            if is_cloud_mode():
                (width, height) = device.display_size()
                Reporter().screenshot_with_mark(width, height, target_pos[0], target_pos[1])

            operator(target_pos[0], target_pos[1], **kwargs)
        else:
            logger.warn("native elem not found : " + str(conditions))
        return target_pos

    @staticmethod
    def get_bound_by_conditions(device, conditions):
        if not device:
            return None
        target_pos = None
        if device.device_type() == DeviceType.DEVICE_ANDROID:
            if device.uiauto_man == None:
                logger.error("handle_element_by_conditions failed : uiauto_man is not set")
                return None
            elem = AutomationHelperNative.__get_uiauto_elem_by_conditions(device, conditions)
            if elem:
                bound_info = elem.info[u'bounds']
                left, top, right, bottom = bound_info[u'left'], bound_info[u'top'], bound_info[u'right'], bound_info[
                    u'bottom']
                return (left, top, right, bottom)

        elif device.device_type() == DeviceType.DEVICE_IOS:
            session = device.wda_session()
            if session is None:
                logger.error("wait_element_by_text failed : session is not set")
                return None
            elem = AutomationHelperNative.__get_ios_elem_by_conditions(device, conditions)
            if elem:
                rect = elem.bound
                return (rect.x, rect.y, rect.x + rect.width, rect.y + rect.height)
        return None


    @staticmethod
    def get_bound_by_self(device, element):
        if not device:
            return None
        if not element:
            return None
        target_pos = None
        if device.device_type() == DeviceType.DEVICE_ANDROID:
            bound_info = element.info[u'bounds']
            left, top, right, bottom = bound_info[u'left'], bound_info[u'top'], bound_info[u'right'], bound_info[
                u'bottom']
            return (left, top, right, bottom)

        elif device.device_type() == DeviceType.DEVICE_IOS:
            rect = element.bound
            return (rect.x, rect.y, rect.x + rect.width, rect.y + rect.height)
        return None


    @staticmethod
    def get_bound_by_text(device, text):
        if not device:
            return None
        target_pos = None
        if device.device_type() == DeviceType.DEVICE_ANDROID:
            if device.uiauto_man == None:
                logger.error("handle_element_by_text failed : uiauto_man is not set")
                return None
            elem = AutomationHelperNative.__get_uiauto_elem_by_text(device, text)
            if elem:
                bound_info = elem.info[u'bounds']
                left, top, right, bottom = bound_info[u'left'], bound_info[u'top'], bound_info[u'right'], bound_info[
                    u'bottom']
                return (left, top, right, bottom)

        elif device.device_type() == DeviceType.DEVICE_IOS:
            session = device.wda_session()
            if session is None:
                logger.error("wait_element_by_text failed : session is not set")
                return None
            elem = AutomationHelperNative.__get_ios_elem_by_text(device, text)
            if elem:
                rect = elem.bound
                return (rect.x, rect.y, rect.x + rect.width, rect.y + rect.height)

        return None



# -*- coding: UTF-8 -*-
import json
import numpy as np
import cv2


def _getPosition(img_query, img_scene, position):
    sceneImgH, sceneImgW = img_scene.shape[:2]
    queryImgH, queryImgW = img_query.shape[:2]
    expand = max(min(queryImgH, queryImgW), 60)
    if sceneImgH > sceneImgW:
        x_min = int(max(position[0] * sceneImgW - expand, 0))
        y_min = int(max(position[1] * sceneImgH - 1.5 * expand, 0))
        x_max = int(min(position[2] * sceneImgW + expand, sceneImgW))
        y_max = int(min(position[3] * sceneImgH + 1.5 * expand, sceneImgH))
    else:
        x_min = int(max(position[0] * sceneImgW - 1.5 * expand, 0))
        y_min = int(max(position[1] * sceneImgH - expand, 0))
        x_max = int(min(position[2] * sceneImgW + 1.5 * expand, sceneImgW))
        y_max = int(min(position[3] * sceneImgH + expand, sceneImgH))
    return (x_min, y_min, x_max, y_max)


def transparent(img_query, img_scene):
    h, w = img_query.shape[:2]
    img_query = cv2.GaussianBlur(img_query, (3, 3), 0)
    img_scene = cv2.GaussianBlur(img_scene, (3, 3), 0)
    binary_threshold = np.mean(img_query)
    img_scene_gray = cv2.cvtColor(img_scene, cv2.COLOR_BGR2GRAY)
    img_query_gray = cv2.cvtColor(img_query, cv2.COLOR_BGR2GRAY)
    retval, binary_scene = cv2.threshold(img_scene_gray, binary_threshold, 255, cv2.THRESH_BINARY)
    retval, binary_query = cv2.threshold(img_query_gray, binary_threshold, 255, cv2.THRESH_BINARY)
    # cv2.imshow('1', binary_query)
    # cv2.waitKey(0)
    kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (2,2))
    binary_query = cv2.erode(binary_query, kernel)
    binary_scene=cv2.erode(binary_scene,kernel)

    res = cv2.matchTemplate(binary_scene, binary_query, cv2.TM_CCOEFF_NORMED)
    _, max_val, _, max_loc = cv2.minMaxLoc(res)
    x_min = max_loc[0]
    y_min = max_loc[1]
    return max_val,(x_min,y_min,w,h)

def highlight(img_query, img_scene):
    img_query_hsv = cv2.cvtColor(img_query, cv2.COLOR_BGR2HSV)
    img_query_v =img_query_hsv[:, :, 2]
    img_query_v_mean = np.mean(img_query_v)
    img_scene_hsv = cv2.cvtColor(img_scene, cv2.COLOR_BGR2HSV)
    img_scene_v = img_scene_hsv[:, :, 2]
    img_scene_v_mean = np.mean(img_scene_v)
    if img_query_v_mean>img_scene_v_mean:
        return True
    else:
        return False

def highlightGraphic(img_query,img_scene,graphic='circle'):


    #find highlight Contours
    h, w = img_scene.shape[:2]
    img_scene_hsv = cv2.cvtColor(img_scene, cv2.COLOR_BGR2HSV)
    img_scene_v = img_scene_hsv[:, :, 2]
    max_v = np.max(img_scene_v)
    img_query_hsv = cv2.cvtColor(img_query, cv2.COLOR_BGR2HSV)
    img_query_v = img_query_hsv[:, :, 2]
    vThreshold = np.mean(img_query_v)
    _, binary = cv2.threshold(img_scene_v, vThreshold, 255, cv2.THRESH_BINARY)
    kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (5, 5))
    binary = cv2.dilate(binary, kernel)
    contours_map, contours, hierarchy = cv2.findContours(binary, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    contours_list = []
    rect_list = []
    arcL=0
    res_cnt=[]
    if len(contours) > 0:
        for i in range(len(contours)):
            cnt = contours[i]
            arcLength = cv2.arcLength(cnt, 1)
            area = cv2.contourArea(cnt)
            if graphic=="circle":
                if arcLength > 80 and area > 250:
                    x, y, w, h = cv2.boundingRect(cnt)
                    if w / h > 0.5 and w / h < 1.6:
                        contours_list.append(cnt)
                        rect_list.append((x, y, w, h))
            if graphic=="rectangle":
                if arcLength > 100:
                    if arcLength > arcL:
                        arcL = arcLength
                        res_cnt = cnt
    # rectangle
    if graphic == "rectangle":
        if res_cnt is None :
            return
        else:
            hull = cv2.convexHull(res_cnt)
            x, y, w, h = cv2.boundingRect(hull)
            # cv2.rectangle(img_scene, (x, y), (x + w, y + h), (0, 255, 0), 1)
            click_x = int(x + w / 2)
            click_y = int(y + h / 2)
            return (click_x, click_y)


    #circle
    if graphic=="circle":
        cnt_id = 0
        click_x = 0
        click_y = 0
        diff = 0.6
        if len(rect_list) > 0:
            for i in range(len(rect_list)):
                x, y, w, h = rect_list[i]
                if abs(w / h - 1) < diff:
                    diff = abs(w / h - 1)
                    cnt_id = i
            rect_x, rect_y, rect_w, rect_h = rect_list[cnt_id]

        circles = cv2.HoughCircles(binary, cv2.HOUGH_GRADIENT, 1, 100, param1=100, param2=30, minRadius=10,
                                   maxRadius=150)

        if circles is None:
            if rect_list:
                click_x = int(rect_x + rect_w / 2)
                click_y = int(rect_y + rect_h / 2)
                return (click_x, click_y)
            else:
                return
        else:
            circles_num = len(circles[0])
            if circles_num == 1:
                circle = circles[0]
                x = int(circle[0][0])
                y = int(circle[0][1])
                r = int(circle[0][2])
                return (x, y)
            elif circles_num > 1:
                res_tag = False

                for circle in circles[0]:
                    x = int(circle[0])
                    y = int(circle[1])
                    r = int(circle[2])

                    if x > rect_x and x < rect_x + rect_w and y > rect_y and y < rect_y + rect_h:
                        res_tag = True

                        return (x, y)
                if not res_tag:
                    return

def judge_SimpleColor(img_query):
    img_query=img_query[:,:,0]
    h, w = img_query.shape[:2]
    hist = cv2.calcHist([img_query], [0], None, [256], [0, 256])
    _, maxVal, _, maxLoc = cv2.minMaxLoc(hist)
    value=maxLoc[1]
    if value<200:
        return False
    else:
        count = 0
        for i in range(value - 5, min(value + 5, 256)):
            count += hist[i]
        percentage=count / h / w
        if percentage>0.6:
            return True
        else:
            return False


def specialBorder(img_scene):
    img_scene_temp0=img_scene[:,:,2].copy()
    img_scene_temp1=~img_scene_temp0
    h,w=img_scene_temp0.shape
    borderWidth=60
    topBorder = 0
    bottomBorder = 0
    leftBorder = 0
    rightBorder = 0
    if h>w:
        for i in range(h):
            if i<borderWidth:
                blackDetect0=np.count_nonzero(img_scene_temp0[i,:]>245)/len(img_scene_temp0[i,:])
                blackDetect1 = np.count_nonzero(img_scene_temp1[i, :] > 245)/len(img_scene_temp0[i,:])
                if blackDetect0>0.9 or blackDetect1>0.9 :
                    topBorder=topBorder+1
                    img_scene[i,:]=(0,0,255)
            if i > h - borderWidth:
                blackDetect0 = np.count_nonzero(img_scene_temp0[i, :] > 245) / len(img_scene_temp0[i, :])
                blackDetect1 = np.count_nonzero(img_scene_temp1[i, :] > 245) / len(img_scene_temp0[i, :])
                if blackDetect0 > 0.9 or blackDetect1 > 0.9:
                    bottomBorder = bottomBorder + 1
                    img_scene[i, :] = (0, 255, 0)


    else:
        for j in range(w):
            if j < borderWidth :
                blackDetect0 = np.count_nonzero(img_scene_temp0[:,j] > 245) / len(img_scene_temp0[:, j])
                blackDetect1 = np.count_nonzero(img_scene_temp1[:, j] > 245) / len(img_scene_temp0[:, j])
                if blackDetect0 > 0.9 or blackDetect1 > 0.9:
                    leftBorder = leftBorder + 1
                    img_scene[:,j] = (0, 0, 255)
            if j < borderWidth or j > w - borderWidth:
                blackDetect0 = np.count_nonzero(img_scene_temp0[:,j] > 245) / len(img_scene_temp0[:, j])
                blackDetect1 = np.count_nonzero(img_scene_temp1[:, j] > 245) / len(img_scene_temp0[:, j])
                if blackDetect0 > 0.9 or blackDetect1 > 0.9:
                    rightBorder = rightBorder + 1
                    img_scene[:,j] = (0, 255, 0)

    if topBorder>40 or bottomBorder>40 or leftBorder>40 or rightBorder>40:
        return True
    else:
        return False


class MatchResult(object):
    """
    匹配结果， 暂时包扩以下字段：
    - ``rect``: 模板位置， (left, top, right, bottom)
    - ``time``: 超找耗时
    - ``scale_ratio``: 得到结果的缩放比例
    - ``parameters``： 请求参数
    - ``miscellaneous``: 其它杂项, 字典类型
    类可以直接用作bool判断， 当rect为非None时返回True

    """
    __slots__ = ["method","rect", "time", "scale_ratio", "parameters", "miscellaneous"]

    def __init__(self, **kwargs):
        self.method=kwargs.get("method",None)
        self.rect = kwargs.get("rect", None)
        self.time = kwargs.get("time", -1)
        self.scale_ratio = kwargs.get("scale_ratio")
        self.parameters = kwargs.get("parameters", {})
        self.miscellaneous = kwargs.get("miscellaneous", {})

    def __bool__(self):
        return self.rect is None

    def __str__(self):
        d = {
            "method":self.method,
            "rect":self.rect,
            "time":self.time,
            "scale_ratio":self.scale_ratio,
            "parameters":self.parameters,
            "miscellaneous":self.miscellaneous
        }
        return json.dumps(d)



import abc, six
from enum import Enum
from ga2.common.loggerConfig import logger
from ga2.common.utils import *
from ga2.engine.engineFactory import EngineFactory
from ga2.cloud import platformHelper
from ga2.engine.engine import *


class DeviceType(Enum):
    DEVICE_ANDROID = 0
    DEVICE_IOS = 1


@six.add_metaclass(abc.ABCMeta)
class Device():
    _engine_connector = None
    _device_sdk_port = 27019
    _local_engine_port = None
    _screenshoter = None
    defaultDevice = None

    @staticmethod
    def get_default():
        return Device.defaultDevice

    @staticmethod
    def set_default(device):
        Device.defaultDevice = device

    def set_default(self):
        Device.defaultDevice = self

    @abc.abstractmethod
    def device_type(self):
        pass

    @abc.abstractmethod
    def push(self, local_path, remote_path):
        pass

    @abc.abstractmethod
    def pull(self, remote_path, local_path):
        pass

    @property
    def screenshoter(self):
        return self._screenshoter

    @screenshoter.setter
    def screenshoter(self, shoter):
        self._screenshoter = shoter

    def engine_connector(self):
        if self._engine_connector is None:
            logger.warn("engine connector is none , try to init it ...")
            self.init_engine_sdk()
        return self._engine_connector

    '''
    init engine gautomator sdk(create a valid client to sdk)
    param:  local_engine_port: a free port used to forward to device engine port
            timeout: engine connecting timeout in seconds
            if running in cloud , a free port will be forward instead of the given param
    '''

    @callLog
    def init_engine_sdk(self, engine_type=None, local_engine_port=None, timeout=60):
        if local_engine_port is not None:
            self._local_engine_port = int(os.environ.get("LOCAL_ENGINE_PORT", local_engine_port))
        self._local_engine_port = self._local_engine_port or 53001
        if is_cloud_mode():
            response = platformHelper.get_platform_client().platform_forward(self._device_sdk_port)
            if response is None:
                return ErrType.ERR_CLOUD_PLATFORM
            else:
                self._local_engine_port = response["localPort"]
        else:
            self.forward(self._local_engine_port, self._device_sdk_port)
        counts = int(timeout / 2 + 1)
        origin_engine_type = engine_type
        for i in range(counts):
            try:
                engine_type = engine_type or "Unity3D"  # use default unity to connect and get the real engine type
                self._engine_connector = EngineFactory.create_engine_connector(engine_type, os.environ.get("PLATFORM_IP") if is_cloud_mode() else "127.0.0.1",
                                                                               int(self._local_engine_port))
                if self._engine_connector is None:
                    logger.error("create engine failed . invalid type : " + str(engine_type))
                    return ErrType.ERR_INVALID_ENGINETYPE
                version = self._engine_connector.get_sdk_version()
                if version:
                    logger.debug(version)
                    if origin_engine_type is None:
                        engine_type = EngineFactory.recognize_engine_type(version)
                        if engine_type is None:
                            logger.error("create engine failed . unrecognized engine type.")
                            return ErrType.ERR_INVALID_ENGINETYPE
                        logger.debug("engine type recognized: " + engine_type)
                        self._engine_connector = EngineFactory.create_engine_connector(engine_type, os.environ.get("PLATFORM_IP") if is_cloud_mode() else "127.0.0.1",
                                                                                       int(self._local_engine_port))
                    return ErrType.ERR_SUCCEED
            except Exception as e:
                time.sleep(2)
        logger.error("init engine sdk timeout")
        return ErrType.ERR_CONNECT_TO_SDK_FAILED

    @abc.abstractmethod
    def forward(self, local_port, remote_port):
        pass

    @abc.abstractmethod
    def remove_forward(self, local_port):
        pass

    @abc.abstractmethod
    def get_top_app(self):
        pass

    @abc.abstractmethod
    def launch_app(self, appid, activity=None, timeout=60, alert_handler=None):
        pass

    def kill_app(self, app):
        pass

    @callLog
    def screenshot(self, localpath=None, portrait=True, height=None):
        return self._screenshoter.screenshot(localpath , portrait , height)

    @abc.abstractmethod
    def display_size(self):
        pass

    @abc.abstractmethod
    def orientation(self):
        pass

    @abc.abstractmethod
    def home(self):
        pass

    @abc.abstractmethod
    def text(self, content):
        pass


    @abc.abstractmethod
    def touch(self, x, y):
        pass



    @abc.abstractmethod
    def double_touch(self, x, y):
        pass

    @abc.abstractmethod
    def long_press(self, x, y, duration=2):
        pass

    @abc.abstractmethod
    def drag(self, sx, sy, dx, dy, duration=1):
        pass

    @abc.abstractmethod
    def clear_app(self, appid):
        logger.info("ios clear_app : unimplemented yet")

    @abc.abstractmethod
    def keyevent(self, keycode):
        logger.info("keyevent : unimplemented yet")



    # '''
    # touch the center of the given rectangle bound.
    # param: bound: a list as [s,y,width,height].
    # '''
    # def __touch_bound(self, bound):
    #     if 0 < bound[0] < 1 and 0 < bound[1] < 1 and 0 < bound[2] < 1 and 0 < bound[3] < 1:
    #         display_size = self.display_size()
    #         bound[0] *= display_size[0]
    #         bound[1] *= display_size[1]
    #         bound[2] *= display_size[0]
    #         bound[3] *= display_size[1]
    #     x, y = bound[0] + bound[2] / 2, bound[1] + bound[3] / 2
    #     self.touch(x, y)
    #
    # def __touch_hold_bound(self, bound, duration=1.5):
    #     if 0 < bound[0] < 1 and 0 < bound[1] < 1 and 0 < bound[2] < 1 and 0 < bound[3] < 1:
    #         display_size = self.display_size()
    #         bound[0] *= display_size[0]
    #         bound[1] *= display_size[1]
    #         bound[2] *= display_size[0]
    #         bound[3] *= display_size[1]
    #     x, y = bound[0] + bound[2] / 2, bound[1] + bound[3] / 2
    #     self.long_press(x, y, duration)








#-*- coding: UTF-8 -*-
__author__ = 'minhuaxu'

class ConnectionException(RuntimeError):
    """

    """
    pass

class HttpException(ConnectionException):
    """
        http请求发生错误
    """
    pass



from enum import Enum


class By(Enum):
    NAME_IN_ENGINE = 100
    PATH_IN_ENGINE = 101
    TEXT_NATIVE = 102
    CONDITIONS_NATIVE = 103
    IMAGE_TEMPLATE = 104
    ID_IN_UICONFIG = 105
    ELEMENT_IN_ENGINE = 106
    ELEMENT_NATIVE = 107


# -*- coding: UTF-8 -*-

from ga2.device.device import Device
from ga2.device.device import DeviceType
from ga2.device.deviceOrientation import DeviceOrientation
from ga2.common.loggerConfig import logger
from ga2.device.android.inputHelper import InputHeler
from ga2.common.utils import *
from ga2.device.android.adbProcess import *
from ga2.device.android.uiauto.uiautoManager import UiautoManager
from ga2.device.android.defaultScreenShoter import DefaultScreenshoter
from ga2.cloud import platformHelper
import atexit


class AndroidDevice(Device):

    def __init__(self, serial, need_uiauto=True, uiautoport=None, gaport=None, screenshoter=None):
        self.uiauto_man = UiautoManager(serial, os.environ.get("PLATFORM_IP", "127.0.0.1"), port=uiautoport) if need_uiauto else None
        self._serial = serial
        self._local_engine_port = gaport
        self._screenshoter = screenshoter or DefaultScreenshoter(serial, self.uiauto_man)
        atexit.register(self.__cleanup)
        self.__forward_list = []
        pass

    def __cleanup(self):  #clean forward list ?
        for port in self.__forward_list:
            self.remove_forward(port)

    def device_type(self):
        return DeviceType.DEVICE_ANDROID

    '''
       :param package:
       :return: errorcode
       :note: app will not be relaunched if not being foreground
    '''
    @callLog
    def launch_app(self, package, activity=None, timeout=60, alert_handler=None):
        activity = activity or os.environ.get("LAUNCHACTIVITY", "android.intent.category.LAUNCHER")
        if is_cloud_mode() :
            if platformHelper.get_platform_client().launch_app(package, activity) is True:
                return ErrType.ERR_SUCCEED
        adb_wait("shell monkey -p " + package + " -c " + activity + " 1", self._serial)
        pid = adb_get_pid_by_name(package)

        return ErrType.ERR_SUCCEED if pid is not None and pid > 0 else ErrType.ERR_PROCESS_NOT_FOUND

    def get_top_app(self):
        package = adb_get_top_app(self._serial)
        if package is None and self.uiauto_man:
            package = self.uiauto_man.ui_device.info["currentPackageName"]
            if package:
                logger.debug("found top package by uiauto = {0}".format(package))
        return package

    '''
    kill app by packagename
    '''
    @callLog
    def kill_app(self, appid=None):
        if appid is not None:
            adb_wait("shell am force-stop " + appid, self._serial)

    '''
    clear app data by packagename
    '''
    def clear_app(self, appid):
        cmd = "shell pm clear {0}".format(appid)
        adb_wait(cmd, self._serial)

    @callLog
    def home(self):
        return self.keyevent(keycode=3)

    def keyevent(self, keycode):
        out, err = adb_wait('shell input keyevent ' + str(keycode), self._serial)
        return ErrType.ERR_SUCCEED if err == "" or err is None else ErrType.ERR_ADB

    def forward(self, local_port, remote_port):
        out, err = adb_forward(local_port, remote_port, self._serial)
        if err is not None and len(err) > 0:
            return ErrType.ERR_ADB
        self.__forward_list.append(local_port)
        return ErrType.ERR_SUCCEED

    def remove_forward(self, local_port):
        out, err = adb_remove_forward(local_port, self._serial)
        if err is not None and len(err) > 0:
            return ErrType.ERR_ADB
        self.__forward_list.remove(local_port)
        return ErrType.ERR_SUCCEED

    def push(self, local_path, remote_path):
        out, err = adb_push(local_path, remote_path, self._serial)
        return ErrType.ERR_SUCCEED if err == "" or err is None else ErrType.ERR_ADB

    def pull(self, remote_path,local_path):
        out, err = adb_pull(remote_path, local_path,self._serial)
        return ErrType.ERR_SUCCEED if err == "" or err is None else ErrType.ERR_ADB

    '''
    input text to device
    param : str
    '''
    def text(self, content):
        InputHeler.input_text(text=content, need_one_by_one=True, serial=self._serial)

    '''
    get the current device display size
    in android , there are two kinds of  display_size (screen display_size and app display_size)
    this method will return the current display size got by adb, which may not  equals to the value got by uiautomator
    return: two values  (width,height)
    '''
    def display_size(self):
        return adb_get_display_size(self._serial)

    '''
    get the current device orientation
    return: enum DeviceOrientation value
    '''
    def orientation(self):
        return DeviceOrientation(self.uiauto_man.ui_device.info["displayRotation"])#0,1,2,3

    '''
    touch at the given position 
    param: x,y : pixel pos  or rate of the current device screen.
    '''
    def touch(self, x, y):
        if 0 < x < 1 and 0 < y < 1:
            display_size = self.display_size()
            x = x * display_size[0]
            y = y * display_size[1]
        adb_wait('shell input tap ' + str(x) + ' ' + str(y), self._serial)

    '''
    double touch at the given position 
    param: x,y : pixel pos  or rate of the current device screen.
    '''
    def double_touch(self, x, y):
        logger.error("double_touch unimplemented yet in AndroidDevice")

    '''
    note:  the param "duration" indicates the moving time(seconds).
    param: sx,sy,dx,dy: start and end pixel pos  or rate of the current device screen.
           duration: ms
    '''
    def drag(self, sx, sy, dx, dy, duration=0.3):
        if 0 < sx < 1 and 0 < sy < 1 and 0 < dx < 1 and 0 < dy < 1:
            display_size = self.display_size()
            sx = sx * display_size[0]
            sy = sy * display_size[1]
            dx = dx * display_size[0]
            dy = dy * display_size[1]
        adb_wait('shell input swipe  ' + str(round(sx, 2)) + ' ' + str(round(sy, 2)) + ' ' + str(round(dx, 2)) + ' ' + str(round(dy, 2)) + ' ' + str(int(duration*1000)), self._serial)

    '''
    touch and hold for duration at the given position
    param : x,y : pixel pos  or rate of the current device screen.
            duration: holding seconds
    '''
    def long_press(self, x, y, duration=1.5):
        if 0 < x < 1 and 0 < y < 1:
            display_size = self.display_size()
            x *= display_size[0]
            y *= display_size[1]
        adb_wait('shell input swipe  ' + str(round(x, 2)) + ' ' + str(round(y, 2)) + ' ' + str(round(x+0.1, 2)) + ' ' + str(round(y+0.1, 2)) + ' ' +  str(int(duration*1000)),

                 self._serial)


    click = touch
    swipe = drag




from ga2.cloud import platformHelper
from ga2.common.loggerConfig import logger
from ga2.common.singleton import singleton

@singleton
class Reporter():
    def __init__(self):
        self.errs = []
        self.image_id = 10000
        self.platform_client = platformHelper.get_platform_client()

    def screenshot(self):
        try:
            self.image_id += 1
            response = self.platform_client.take_screenshot()
            if response:
                return response
        except Exception as e:
            logger.exception(e)

    def screenshot_with_mark( self, width, height, x, y, locator_name="point"):
        try:
            self.image_id += 1
            if 0<x<1 and 0<y<1:
                x=int(x*width)
                y=int(y*height)
            response = self.platform_client.touch_capture(width, height, x, y, locator_name)
            if response:
                return response
        except Exception as e:
            logger.exception(e)
